@@ -1,15 +1,15 @@
 use std::io::Write;
+use std::num::NonZeroUsize;
 
+use crate::cmd::parallel::Args as ParallelArgs;
 use crate::config::{Config, Delimiter};
+use crate::moonblade::{
+    AggregationProgram, GroupAggregationProgram, GroupAlongColumnsAggregationProgram,
+};
 use crate::select::SelectColumns;
 use crate::util;
 use crate::CliResult;
 
-use crate::moonblade::AggregationProgram;
-use crate::moonblade::GroupAggregationProgram;
-
-use crate::cmd::moonblade::MoonbladeErrorPolicy;
-
 fn write_group(
     wtr: &mut csv::Writer<Box<dyn Write + Send>>,
     group: &Vec<Vec<u8>>,
@@ -28,6 +28,8 @@ static USAGE: &str = "
 Group a CSV file by values contained in a column selection then aggregate data per
 group using a custom aggregation expression.
 
+For ungrouped aggregation, check the `xan agg` command instead.
+
 The result of running the command will be a CSV file containing the grouped
 columns and additional columns for each computed aggregation.
 
@@ -51,30 +53,63 @@ You can group on multiple columns (read `xan select -h` for more information abo
 
     $ xan groupby name,surname 'sum(count)' file.csv
 
+---
+
+This command is also able to aggregate along columns that you can select using
+the --along-cols <cols> flag. In which case, the aggregation functions will accept
+the anonymous `_` placeholder representing currently processed column's value.
+
+For instance, given the following file:
+
+user,count1,count2
+marcy,4,5
+john,0,1
+marcy,6,8
+john,4,6
+
+Using the following command:
+
+    $ xan groupby user --along-cols count1,count2 'sum(cell)' file.csv
+
+Will produce the following result:
+
+user,count1,count2
+marcy,10,13
+john,4,7
+
+---
+
 For a quick review of the capabilities of the expression language,
 check out the `xan help cheatsheet` command.
 
+For a list of available aggregation functions, use `xan help aggs`.
+
 For a list of available functions, use `xan help functions`.
 
-For a list of available aggregation functions, use `xan help aggs`
-instead.
+Aggregations can be computed in parallel using the -p/--parallel or -t/--threads flags.
+But this cannot work on streams or gzipped files, unless a `.gzi` index (as created
+by `bgzip -i`) can be found beside it. Parallelization is not compatible
+with the -S/--sorted nor -C/--along-cols flags.
 
 Usage:
     xan groupby [options] <column> <expression> [<input>]
     xan groupby --help
 
 groupby options:
-    --keep <cols>           Keep this selection of columns, in addition to
-                            the ones representing groups, in the output. Only
-                            values from the first seen row per group will be kept.
-    -S, --sorted            Use this flag to indicate that the file is already sorted on the
-                            group columns, in which case the command will be able to considerably
-                            optimize memory usage.
-    -e, --errors <policy>   What to do with evaluation errors. One of:
-                              - \"panic\": exit on first error
-                              - \"ignore\": ignore row altogether
-                              - \"log\": print error to stderr
-                            [default: panic].
+    --keep <cols>            Keep this selection of columns, in addition to
+                             the ones representing groups, in the output. Only
+                             values from the first seen row per group will be kept.
+    -C, --along-cols <cols>  Perform a single aggregation over all of selected columns
+                             and create a column per group with the result in the output.
+    -S, --sorted             Use this flag to indicate that the file is already sorted on the
+                             group columns, in which case the command will be able to considerably
+                             optimize memory usage.
+    -p, --parallel           Whether to use parallelization to speed up computation.
+                             Will automatically select a suitable number of threads to use
+                             based on your number of cores. Use -t, --threads if you want to
+                             indicate the number of threads yourself.
+    -t, --threads <threads>  Parellize computations using this many threads. Use -p, --parallel
+                             if you want the number of threads to be automatically chosen instead.
 
 Common options:
     -h, --help               Display this message
@@ -94,14 +129,36 @@ struct Args {
     flag_output: Option<String>,
     flag_delimiter: Option<Delimiter>,
     flag_keep: Option<SelectColumns>,
+    flag_along_cols: Option<SelectColumns>,
     flag_sorted: bool,
-    flag_errors: String,
+    flag_parallel: bool,
+    flag_threads: Option<NonZeroUsize>,
 }
 
 pub fn run(argv: &[&str]) -> CliResult<()> {
     let mut args: Args = util::get_args(USAGE, argv)?;
 
-    let error_policy = MoonbladeErrorPolicy::try_from_restricted(&args.flag_errors)?;
+    if args.flag_parallel || args.flag_threads.is_some() {
+        if args.flag_along_cols.is_some() {
+            Err("-p/--parallel or -t/--threads cannot be used with --along-cols!")?;
+        }
+
+        if args.flag_sorted {
+            Err("-p/--parallel or -t/--threads cannot be used with --sorted!")?;
+        }
+
+        let mut parallel_args = ParallelArgs::single_file(&args.arg_input, args.flag_threads)?;
+
+        parallel_args.cmd_groupby = true;
+        parallel_args.arg_group = Some(args.arg_column);
+        parallel_args.arg_expr = Some(args.arg_expression);
+
+        parallel_args.flag_no_headers = args.flag_no_headers;
+        parallel_args.flag_output = args.flag_output;
+        parallel_args.flag_delimiter = args.flag_delimiter;
+
+        return parallel_args.run();
+    }
 
     let rconf = Config::new(&args.arg_input)
         .delimiter(args.flag_delimiter)
@@ -114,10 +171,55 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
 
     let sel = rconf.selection(headers)?;
 
-    // Lol, what a hack...
+    // --along-cols
+    if let Some(selection) = args.flag_along_cols.take() {
+        if args.flag_sorted || args.flag_keep.is_some() {
+            Err("--along-cols does not work with -S/--sorted nor --keep!")?;
+        }
+
+        let mut pivot_sel = selection.selection(headers, !args.flag_no_headers)?;
+        pivot_sel.sort_and_dedup();
+
+        let mut program = GroupAlongColumnsAggregationProgram::parse(
+            &args.arg_expression,
+            headers,
+            pivot_sel.len(),
+        )?;
+
+        if !args.flag_no_headers {
+            let mut output_headers = sel.select(headers).collect::<csv::ByteRecord>();
+
+            for name in pivot_sel.select(headers) {
+                output_headers.push_field(name);
+            }
+
+            wtr.write_byte_record(&output_headers)?;
+        }
+
+        let mut record = csv::ByteRecord::new();
+        let mut index: usize = 0;
+
+        while rdr.read_byte_record(&mut record)? {
+            let group = sel.collect(&record);
+
+            program.run_with_cells(group, index, &record, pivot_sel.select(&record))?;
+
+            index += 1;
+        }
+
+        for result in program.into_byte_records(false) {
+            let (group, group_record) = result?;
+
+            write_group(&mut wtr, &group, &group_record)?;
+        }
+
+        return Ok(wtr.flush()?);
+    }
+
+    // --keep, lol...
     if let Some(selection) = args.flag_keep.take() {
         let mut keep_sel = selection.selection(headers, !args.flag_no_headers)?;
-        keep_sel.sort_and_dedup();
+        keep_sel.dedup();
 
         let addendum = keep_sel
             .iter()
@@ -144,11 +246,13 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
         let mut program = AggregationProgram::parse(&args.arg_expression, headers)?;
         let mut current: Option<Vec<Vec<u8>>> = None;
 
-        write_group(
-            &mut wtr,
-            &sel.collect(headers),
-            &program.headers().collect(),
-        )?;
+        if !args.flag_no_headers {
+            write_group(
+                &mut wtr,
+                &sel.collect(headers),
+                &program.headers().collect(),
+            )?;
+        }
 
         let mut index: usize = 0;
 
@@ -161,55 +265,45 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
                 }
                 Some(current_group) => {
                     if current_group != &group {
-                        write_group(
-                            &mut wtr,
-                            current_group,
-                            &error_policy.handle_error(program.finalize(false))?,
-                        )?;
+                        write_group(&mut wtr, current_group, &program.finalize(false)?)?;
                         program.clear();
                         current = Some(group);
                     }
                 }
             };
 
-            program
-                .run_with_record(index, &record)
-                .or_else(|error| error_policy.handle_row_error(index, error))?;
+            program.run_with_record(index, &record)?;
 
             index += 1;
         }
 
         // Flushing final group
         if let Some(current_group) = current {
-            write_group(
-                &mut wtr,
-                &current_group,
-                &error_policy.handle_error(program.finalize(false))?,
-            )?;
+            write_group(&mut wtr, &current_group, &program.finalize(false)?)?;
         }
     } else {
         let mut program = GroupAggregationProgram::parse(&args.arg_expression, headers)?;
 
-        write_group(
-            &mut wtr,
-            &sel.collect(headers),
-            &program.headers().collect(),
-        )?;
+        if !args.flag_no_headers {
+            write_group(
+                &mut wtr,
+                &sel.collect(headers),
+                &program.headers().collect(),
+            )?;
+        }
 
         let mut index: usize = 0;
 
         while rdr.read_byte_record(&mut record)? {
             let group = sel.collect(&record);
 
-            program
-                .run_with_record(group, index, &record)
-                .or_else(|error| error_policy.handle_row_error(index, error))?;
+            program.run_with_record(group, index, &record)?;
 
             index += 1;
         }
 
         for result in program.into_byte_records(false) {
-            let (group, group_record) = error_policy.handle_error(result)?;
+            let (group, group_record) = result?;
 
             write_group(&mut wtr, &group, &group_record)?;
         }
