@@ -1,14 +1,13 @@
 // Issue tracking:
 //  - https://github.com/ratatui/ratatui/issues/334
 //  - https://github.com/ratatui/ratatui/issues/1391
-
-use std::collections::HashMap;
 use std::convert::TryFrom;
 use std::num::NonZeroUsize;
 
+use ahash::RandomState;
 use indexmap::IndexMap;
 use jiff::{
-    civil::{Date, DateTime, Time},
+    civil::{Date, Time},
     tz::TimeZone,
     Timestamp, Unit, Zoned, ZonedRound,
 };
@@ -20,8 +19,9 @@ use ratatui::style::{Style, Stylize};
 use ratatui::symbols;
 use ratatui::widgets::{Axis, Chart, Dataset, GraphType};
 
+use crate::collections::HashMap;
 use crate::config::{Config, Delimiter};
-use crate::dates::{infer_temporal_granularity, parse_partial_date};
+use crate::dates::{infer_temporal_granularity, parse_partial_date, parse_zoned};
 use crate::ratatui::print_ratatui_frame_to_stdout;
 use crate::scales::{Scale, ScaleType};
 use crate::select::SelectColumns;
@@ -278,6 +278,18 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
     let (flag_x_min, flag_x_max) = args.parse_x_bounds()?;
     let (flag_y_min, flag_y_max) = (args.flag_y_min, args.flag_y_max);
 
+    if args.flag_x_scale.is_logarithmic()
+        && (matches!(flag_x_min, Some(v) if v <= 0.0) || matches!(flag_x_max, Some(v) if v <= 0.0))
+    {
+        Err("--x-min or --x-max cannot be <= 0 with --x-scale log!")?;
+    }
+
+    if args.flag_y_scale.is_logarithmic()
+        && (matches!(flag_y_min, Some(v) if v <= 0.0) || matches!(flag_y_max, Some(v) if v <= 0.0))
+    {
+        Err("--y-min or --y-max cannot be <= 0 with --y-scale log!")?;
+    }
+
     if args.flag_category.is_some() && !args.flag_add_series.is_empty() {
         Err("-c, --category cannot work with -Y, --add-series!")?;
     }
@@ -784,12 +796,8 @@ fn parse_as_timestamp(cell: &[u8]) -> Result<f64, CliError> {
 
     let string = std::str::from_utf8(cell).map_err(|_| format_error())?;
 
-    let zoned = if let Ok(z) = string.parse::<Zoned>() {
+    let zoned = if let Ok(z) = parse_zoned(string, None, None) {
         z
-    } else if let Ok(datetime) = string.parse::<DateTime>() {
-        datetime
-            .to_zoned(TimeZone::system())
-            .map_err(|_| format_error())?
     } else if let Ok(date) = string.parse::<Date>() {
         date.to_datetime(Time::default())
             .to_zoned(TimeZone::system())
@@ -1296,8 +1304,10 @@ impl Series {
         self.regression_line().map(|(intercept, slope)| {
             let (min_x, max_x) = self.extent.unwrap().0;
 
-            let first_point = (0.0, scales.1.percent(intercept + slope * min_x));
-            let second_point = (1.0, scales.1.percent(intercept + slope * max_x));
+            let mut first_point = (0.0, scales.1.percent(intercept + slope * min_x));
+            let mut second_point = (1.0, scales.1.percent(intercept + slope * max_x));
+
+            clip(&mut first_point, &mut second_point, [0.0, 0.0, 1.0, 1.0]);
 
             [first_point, second_point]
         })
@@ -1306,7 +1316,7 @@ impl Series {
 
 #[derive(Default)]
 struct CategoricalSeries {
-    mapping: IndexMap<Vec<u8>, Series>,
+    mapping: IndexMap<Vec<u8>, Series, RandomState>,
 }
 
 impl CategoricalSeries {
@@ -1420,3 +1430,79 @@ impl SeriesBuilder {
         }
     }
 }
+
+fn clip_t(n: f64, d: f64, c: &mut (f64, f64)) -> bool {
+    let t_e = c.0;
+    let t_l = c.1;
+
+    if d.abs() < f64::EPSILON {
+        return n < 0.0;
+    }
+
+    let t = n / d;
+
+    if d > 0.0 {
+        if t > t_l {
+            return false;
+        }
+        if t > t_e {
+            c.0 = t;
+        }
+    } else {
+        if t < t_e {
+            return false;
+        }
+        if t < t_l {
+            c.1 = t;
+        }
+    }
+
+    true
+}
+
+// NOTE: this is an implementation of Liang-Barsky clipping
+// NOTE: true means inside
+// NOTE: bb is [xmin, ymin, xmax, ymax]
+fn clip(a: &mut (f64, f64), b: &mut (f64, f64), bb: [f64; 4]) -> bool {
+    let x1 = a.0;
+    let y1 = a.1;
+    let x2 = b.0;
+    let y2 = b.1;
+
+    let dx = x2 - x1;
+    let dy = y2 - y1;
+
+    if dx.abs() < f64::EPSILON
+        && dy.abs() < f64::EPSILON
+        && x1 >= bb[0]
+        && x1 <= bb[2]
+        && y1 >= bb[1]
+        && y1 <= bb[3]
+    {
+        return true;
+    }
+
+    let mut c = (0.0, 1.0);
+
+    if clip_t(bb[0] - x1, dx, &mut c)
+        && clip_t(x1 - bb[2], -dx, &mut c)
+        && clip_t(bb[1] - y1, dy, &mut c)
+        && clip_t(y1 - bb[3], -dy, &mut c)
+    {
+        let t_e = c.0;
+        let t_l = c.1;
+
+        if t_l < 1.0 {
+            b.0 = x1 + t_l * dx;
+            b.1 = y1 + t_l * dy;
+        }
+        if t_e > 0.0 {
+            a.0 += t_e * dx;
+            a.1 += t_e * dy;
+        }
+
+        return true;
+    }
+
+    false
+}
