@@ -4,16 +4,16 @@ use std::borrow::{Borrow, ToOwned};
 use std::convert::TryFrom;
 use std::env;
 use std::fs;
-use std::io::{self, prelude::*, BufReader, IsTerminal, Read};
+use std::io::{self, prelude::*, BufReader, IsTerminal, Read, SeekFrom};
 use std::ops::Deref;
-use std::path::PathBuf;
+use std::path::{Path, PathBuf};
 
-use crate::index::Indexed;
-use flate2::read::GzDecoder;
+use bgzip::index::BGZFIndex;
+use bgzip::read::{BGZFReader, IndexedBGZFReader};
+use flate2::read::MultiGzDecoder;
 
-use crate::read::ReverseRead;
+use crate::read::{self, ReverseRead};
 use crate::select::{SelectColumns, Selection};
-use crate::util;
 use crate::{CliError, CliResult};
 
 #[derive(Clone, Copy, Debug, Deserialize)]
@@ -62,6 +62,18 @@ impl TryFrom<String> for Delimiter {
     }
 }
 
+#[derive(Debug)]
+enum TabularDataKind {
+    RegularCsv,
+    Cdx,
+}
+
+impl TabularDataKind {
+    fn is_cdx(&self) -> bool {
+        matches!(self, Self::Cdx)
+    }
+}
+
 pub trait SeekRead: Seek + Read {}
 impl<T: Seek + Read> SeekRead for T {}
 
@@ -70,7 +82,6 @@ type PairResult = CliResult<(String, Option<String>)>;
 #[derive(Debug)]
 pub struct Config {
     pub path: Option<PathBuf>, // None implies <stdin>
-    idx_path: Option<PathBuf>,
     select_columns: Option<SelectColumns>,
     delimiter: u8,
     pub no_headers: bool,
@@ -81,15 +92,18 @@ pub struct Config {
     double_quote: bool,
     escape: Option<u8>,
     quoting: bool,
+    compressed: bool, // TODO: can become a compression type if we need to support more schemes than gz
+    tabular_data_kind: TabularDataKind,
 }
 
 impl Config {
     pub fn new(path: &Option<String>) -> Config {
-        let (path, delim) = match *path {
-            None => (None, b','),
-            Some(ref s) if s.deref() == "-" => (None, b','),
+        let (path, delim, compressed, tabular_data_kind) = match *path {
+            None => (None, b',', false, TabularDataKind::RegularCsv),
+            Some(ref s) if s.deref() == "-" => (None, b',', false, TabularDataKind::RegularCsv),
             Some(ref s) => {
                 let raw_s = s.strip_suffix(".gz").unwrap_or(s);
+                let mut kind = TabularDataKind::RegularCsv;
 
                 let delim = if raw_s.ends_with(".tsv") || raw_s.ends_with(".tab") {
                     b'\t'
@@ -97,16 +111,19 @@ impl Config {
                     b';'
                 } else if raw_s.ends_with(".psv") {
                     b'|'
+                } else if raw_s.ends_with(".cdx") {
+                    kind = TabularDataKind::Cdx;
+                    b' '
                 } else {
                     b','
                 };
 
-                (Some(PathBuf::from(s)), delim)
+                (Some(PathBuf::from(s)), delim, s.ends_with(".gz"), kind)
             }
         };
-        Config {
+
+        let mut config = Config {
             path,
-            idx_path: None,
             select_columns: None,
             delimiter: delim,
             no_headers: false,
@@ -117,10 +134,18 @@ impl Config {
             double_quote: true,
             escape: None,
             quoting: true,
+            compressed,
+            tabular_data_kind,
+        };
+
+        if config.tabular_data_kind.is_cdx() {
+            config.quoting = false;
         }
+
+        config
     }
 
-    pub fn empty() -> Config {
+    pub fn stdin() -> Config {
         Self::new(&None)
     }
 
@@ -235,106 +260,52 @@ impl Config {
         Ok(self.csv_writer_from_writer(self.io_writer_with_options(options)?))
     }
 
-    pub fn reader(&self) -> io::Result<csv::Reader<Box<dyn io::Read + Send + 'static>>> {
+    #[allow(clippy::single_match)]
+    fn read_typical_headers<R: Read>(&self, reader: &mut R) -> CliResult<()> {
+        match self.tabular_data_kind {
+            TabularDataKind::Cdx => {
+                if !read::consume_cdx_header(reader)? {
+                    Err("invalid CDX header!")?;
+                }
+            }
+            _ => (),
+        };
+
+        Ok(())
+    }
+
+    pub fn reader(&self) -> CliResult<csv::Reader<Box<dyn io::Read + Send + 'static>>> {
         Ok(self.csv_reader_from_reader(self.io_reader()?))
     }
 
-    pub fn reader_file(&self) -> io::Result<csv::Reader<Box<dyn SeekRead>>> {
-        self.io_reader_for_random_access_with_cursor_fallback()
-            .map(|f| self.csv_reader_from_reader(f))
+    pub fn seekable_reader(&self) -> CliResult<csv::Reader<Box<dyn SeekRead + Send + 'static>>> {
+        Ok(self.csv_reader_from_reader(self.io_reader_for_random_access()?))
     }
 
-    pub fn index_files(&self) -> io::Result<Option<(csv::Reader<fs::File>, fs::File)>> {
-        let (csv_file, idx_file) = match (&self.path, &self.idx_path) {
-            (&None, &None) => return Ok(None),
-            (&None, &Some(_)) => {
-                return Err(io::Error::new(
-                    io::ErrorKind::Other,
-                    "Cannot use <stdin> with indexes",
-                    // Some(format!("index file: {}", p.display()))
-                ));
-            }
-            (Some(p), &None) => {
-                // We generally don't want to report an error here, since we're
-                // passively trying to find an index.
-                let idx_file = match fs::File::open(util::idx_path(p)) {
-                    // TODO: Maybe we should report an error if the file exists
-                    // but is not readable.
-                    Err(_) => return Ok(None),
-                    Ok(f) => f,
-                };
-                (fs::File::open(p)?, idx_file)
-            }
-            (Some(p), Some(ip)) => (fs::File::open(p)?, fs::File::open(ip)?),
-        };
-        // If the CSV data was last modified after the index file was last
-        // modified, then return an error and demand the user regenerate the
-        // index.
-        let data_modified = util::last_modified(&csv_file.metadata()?);
-        let idx_modified = util::last_modified(&idx_file.metadata()?);
-        if data_modified > idx_modified {
-            return Err(io::Error::new(
-                io::ErrorKind::Other,
-                "The CSV file was modified after the index file. \
-                 Please re-create the index.",
-            ));
-        }
-        let csv_rdr = self.csv_reader_from_reader(csv_file);
-        Ok(Some((csv_rdr, idx_file)))
-    }
-
-    pub fn indexed(&self) -> CliResult<Option<Indexed<fs::File, fs::File>>> {
-        match self.index_files()? {
-            None => Ok(None),
-            Some((r, i)) => Ok(Some(Indexed::open(r, i)?)),
-        }
-    }
-
-    pub fn io_reader(&self) -> io::Result<Box<dyn io::Read + Send + 'static>> {
+    pub fn io_reader(&self) -> CliResult<Box<dyn io::Read + Send + 'static>> {
         Ok(match self.path {
             None => {
                 if io::stdin().is_terminal() {
-                    return Err(io::Error::new(io::ErrorKind::NotFound, "failed to read CSV data from stdin. Did you forget to give a path to your file?"));
+                    return Err(io::Error::new(io::ErrorKind::NotFound, "failed to read CSV data from stdin. Did you forget to give a path to your file?"))?;
                 } else {
                     Box::new(io::stdin())
                 }
             }
             Some(ref p) => match fs::File::open(p) {
                 Ok(x) => {
-                    if p.to_string_lossy().ends_with(".gz") {
-                        Box::new(GzDecoder::new(x))
+                    let mut reader: Box<dyn Read + Send + 'static> = if self.compressed {
+                        Box::new(MultiGzDecoder::new(x))
                     } else {
                         Box::new(x)
-                    }
-                }
-                Err(err) => {
-                    let msg = format!("failed to open {}: {}", p.display(), err);
-                    return Err(io::Error::new(io::ErrorKind::NotFound, msg));
-                }
-            },
-        })
-    }
+                    };
 
-    pub fn io_buf_reader(&self) -> io::Result<Box<dyn io::BufRead + Send + 'static>> {
-        Ok(match self.path {
-            None => {
-                if io::stdin().is_terminal() {
-                    return Err(io::Error::new(io::ErrorKind::NotFound, "failed to read CSV data from stdin. Did you forget to give a path to your file?"));
-                } else {
-                    Box::new(BufReader::new(io::stdin()))
-                }
-            }
-            Some(ref p) => match fs::File::open(p) {
-                Ok(x) => {
-                    if p.to_string_lossy().ends_with(".gz") {
-                        Box::new(BufReader::new(GzDecoder::new(x)))
-                    } else {
-                        Box::new(BufReader::new(x))
-                    }
+                    self.read_typical_headers(&mut reader)?;
+
+                    reader
                 }
                 Err(err) => {
                     let msg = format!("failed to open {}: {}", p.display(), err);
-                    return Err(io::Error::new(io::ErrorKind::NotFound, msg));
+                    return Err(io::Error::new(io::ErrorKind::NotFound, msg))?;
                 }
             },
         })
@@ -362,7 +333,7 @@ impl Config {
             )));
         }
 
-        let lines_reader = self.io_buf_reader()?;
+        let lines_reader = BufReader::new(self.io_reader()?);
 
         Ok(Box::new(lines_reader.lines().filter_map(
             |result| match result {
@@ -412,7 +383,7 @@ impl Config {
             )));
         }
 
-        let lines_reader = self.io_buf_reader()?;
+        let lines_reader = BufReader::new(self.io_reader()?);
 
         Ok(Box::new(lines_reader.lines().filter_map(
             |result| match result {
@@ -430,81 +401,116 @@ impl Config {
         )))
     }
 
-    pub fn io_reader_for_random_access(&self) -> io::Result<Box<dyn SeekRead + 'static>> {
-        let msg = "can't use provided input because it does not allow for random access (e.g. stdin or piping)".to_string();
+    pub fn is_indexed_gzip(&self) -> bool {
         match self.path {
-            None => Err(io::Error::new(io::ErrorKind::Unsupported, msg)),
-            Some(ref p) => match fs::File::open(p) {
-                Ok(x) => match x.borrow().stream_position() {
-                    Ok(_) => Ok(Box::new(x)),
-                    Err(_) => Err(io::Error::new(io::ErrorKind::Unsupported, msg)),
-                },
-                Err(err) => {
-                    let msg = format!("failed to open {}: {}", p.display(), err);
-                    Err(io::Error::new(io::ErrorKind::NotFound, msg))
+            None => false,
+            Some(ref p) => {
+                if self.compressed {
+                    let index_path_str = p.to_string_lossy() + ".gzi";
+                    let index_path = Path::new(index_path_str.as_ref());
+
+                    index_path.is_file()
+                } else {
+                    false
                 }
-            },
+            }
         }
     }
 
-    pub fn io_reader_for_random_access_with_cursor_fallback(
-        &self,
-    ) -> io::Result<Box<dyn SeekRead + 'static>> {
+    pub fn io_reader_for_random_access(&self) -> CliResult<Box<dyn SeekRead + Send + 'static>> {
+        let msg = "can't use provided input because it does not allow for random access (e.g. stdin or piping)".to_string();
+
         match self.path {
-            None => Ok(Box::new(util::bytes_cursor_from_read(&mut io::stdin())?)),
+            None => Err(io::Error::new(io::ErrorKind::Unsupported, msg))?,
             Some(ref p) => match fs::File::open(p) {
                 Ok(mut x) => {
-                    if p.to_string_lossy().ends_with(".gz") {
-                        Ok(Box::new(util::bytes_cursor_from_read(
-                            &mut GzDecoder::new(x),
-                        )?))
-                    } else {
-                        match x.borrow().stream_position() {
-                            Ok(_) => Ok(Box::new(x)),
-                            Err(_) => Ok(Box::new(util::bytes_cursor_from_read(&mut x)?)),
+                    if self.compressed {
+                        let index_path_str = p.to_string_lossy() + ".gzi";
+                        let index_path = Path::new(index_path_str.as_ref());
+
+                        if index_path.is_file() {
+                            let reader = BGZFReader::new(x)?;
+                            let index = BGZFIndex::from_reader(fs::File::open(index_path)?)?;
+                            let mut indexed_reader = IndexedBGZFReader::new(reader, index)?;
+
+                            self.read_typical_headers(&mut indexed_reader)?;
+
+                            return Ok(Box::new(indexed_reader));
                         }
+                    } else {
+                        self.read_typical_headers(&mut x)?;
+                    }
+
+                    match x.borrow().stream_position() {
+                        Ok(_) => Ok(Box::new(x)),
+                        Err(_) => Err(io::Error::new(io::ErrorKind::Unsupported, msg))?,
                     }
                 }
                 Err(err) => {
                     let msg = format!("failed to open {}: {}", p.display(), err);
-                    Err(io::Error::new(io::ErrorKind::NotFound, msg))
+                    Err(io::Error::new(io::ErrorKind::NotFound, msg))?
                 }
             },
         }
     }
 
-    pub fn io_reader_for_reverse_reading(
+    pub fn io_reader_at_position(
         &self,
-        offset: u64,
-    ) -> io::Result<Box<dyn io::Read + 'static>> {
-        let msg = "can't use provided input because it does not allow for random access (e.g. stdin or piping)".to_string();
-        match self.path {
-            None => Err(io::Error::new(io::ErrorKind::Unsupported, msg)),
-            Some(ref p) => match fs::File::open(p) {
-                Ok(x) => match x.borrow().stream_position() {
-                    Ok(_) => {
-                        let filesize = x.metadata()?.len();
-                        Ok(Box::new(ReverseRead::new(x, filesize, offset)))
-                    }
-                    Err(_) => Err(io::Error::new(io::ErrorKind::Unsupported, msg)),
-                },
-                Err(err) => {
-                    let msg = format!("failed to open {}: {}", p.display(), err);
-                    Err(io::Error::new(io::ErrorKind::NotFound, msg))
-                }
-            },
-        }
+        position: u64,
+    ) -> CliResult<Box<dyn Read + Send + 'static>> {
+        let mut reader = self.io_reader_for_random_access()?;
+
+        reader.seek(SeekFrom::Start(position))?;
+
+        Ok(Box::new(reader))
     }
 
-    pub fn csv_reader_from_reader<R: Read>(&self, rdr: R) -> csv::Reader<R> {
-        csv::ReaderBuilder::new()
+    pub fn reverse_reader(
+        &self,
+    ) -> CliResult<(
+        csv::ByteRecord,
+        csv::Reader<Box<dyn io::Read + Send + 'static>>,
+    )> {
+        let mut io_reader = self.io_reader_for_random_access()?;
+        let offset_before_csv_parsing = io_reader.stream_position()?;
+
+        let mut forward_reader = self.csv_reader_from_reader(io_reader);
+        let headers = forward_reader.byte_headers()?.clone();
+
+        let offset = if self.no_headers {
+            offset_before_csv_parsing
+        } else {
+            offset_before_csv_parsing + forward_reader.position().byte()
+        };
+
+        let filesize = forward_reader.get_mut().seek(SeekFrom::End(0))?;
+
+        let reverse_reader = ReverseRead::new(forward_reader.into_inner(), filesize, offset);
+        let mut reader_builder = self.csv_reader_builder();
+        reader_builder.has_headers(false);
+        dbg!(offset, filesize);
+        Ok((
+            headers,
+            reader_builder.from_reader(Box::new(reverse_reader)),
+        ))
+    }
+
+    pub fn csv_reader_builder(&self) -> csv::ReaderBuilder {
+        let mut builder = csv::ReaderBuilder::new();
+
+        builder
             .flexible(self.flexible)
             .delimiter(self.delimiter)
             .has_headers(!self.no_headers)
             .quote(self.quote)
             .quoting(self.quoting)
-            .escape(self.escape)
-            .from_reader(rdr)
+            .escape(self.escape);
+
+        builder
+    }
+
+    pub fn csv_reader_from_reader<R: Read>(&self, rdr: R) -> csv::Reader<R> {
+        self.csv_reader_builder().from_reader(rdr)
     }
 
     fn io_writer_with_options(
