@@ -15,7 +15,6 @@ mod collections;
 mod config;
 mod dates;
 mod graph;
-mod index;
 mod json;
 mod moonblade;
 mod ratatui;
@@ -45,6 +44,8 @@ macro_rules! command_list {
 ## Search & filter
     search      Search for (or replace) patterns in CSV data
     filter      Only keep some CSV rows based on an evaluated expression
+    head        First rows of CSV file
+    tail        Last rows of CSV file
     slice       Slice rows of CSV file
     top         Find top rows of a CSV file according to some column
     sample      Randomly sample CSV data
@@ -60,12 +61,12 @@ macro_rules! command_list {
     stats            Compute basic statistics
     agg              Aggregate data from CSV file
     bins             Dispatch numeric columns into bins
+    window           Compute window aggregations (cumsum, rolling mean, lag etc.)
 
 ## Combine multiple CSV files
     cat         Concatenate by row or column
     join        Join CSV files
-    regex-join  Fuzzy join CSV files using regex patterns
-    url-join    Join CSV files on url prefixes
+    fuzzy-join  Join a CSV file with another containing patterns (e.g. regexes)
     merge       Merge multiple similar already sorted CSV files
 
 ## Add, transform, drop and move columns
@@ -91,21 +92,19 @@ macro_rules! command_list {
     scrape        Scrape HTML into CSV data
     reverse       Reverse rows of CSV data
     transpose (t) Transpose CSV file
+    pivot         Stack multiple columns into fewer columns
+    unpivot       Split distinct values of a column into their own columns
 
 ## Split a CSV file into multiple
     split       Split CSV data into chunks
     partition   Partition CSV data based on a column value
 
-## Parallel operation over multiple CSV files
-    parallel (p) Map-reduce-like parallel computation over multiple CSV files
+## Parallelization
+    parallel (p) Map-reduce-like parallel computation
 
 ## Generate CSV files
     range       Create a CSV file from a numerical range
 
-## Perform side-effects
-    eval        Evaluate/debug a single expression
-    foreach     Loop over a CSV file to perform side effects
-
 ## Lexicometry & fuzzy matching
     tokenize    Tokenize a text column
     vocab       Build a vocabulary over tokenized documents
@@ -114,6 +113,9 @@ macro_rules! command_list {
 ## Matrix & network-related commands
     matrix      Convert CSV data to matrix data
     network     Convert CSV data to network data
+
+## Debug
+    eval  Evaluate/debug a single expression
 "
     };
 }
@@ -121,7 +123,7 @@ macro_rules! command_list {
 static USAGE: &str = concat!(
     "
 Usage:
-    xan <command> [<args>...]
+    xan [options] <command> [<args>...]
     xan [options]
 
 Options:
@@ -129,6 +131,10 @@ Options:
     <command> -h  Display the command help message
     --version     Print version info and exit
 
+Common options:
+    -n, --no-headers       Typically used to indicate that input file has no headers.
+    -d, --delimiter <arg>  Typically used to indicate a custom delimiter.
+
 Commands:",
     command_list!()
 );
@@ -138,7 +144,18 @@ struct Args {
     arg_command: Option<Command>,
 }
 
+// Ref: https://github.com/medialab/xan/issues/566
+#[cfg(not(windows))]
+fn set_virtual_terminal() {}
+
+#[cfg(windows)]
+fn set_virtual_terminal() {
+    colored::control::set_virtual_terminal(true).ok();
+}
+
 fn main() {
+    set_virtual_terminal();
+
     let args: Args = Docopt::new(USAGE)
         .and_then(|d| {
             d.options_first(true)
@@ -189,24 +206,35 @@ Please choose one of the following commands/flags:\n{}",
             Ok(()) => process::exit(0),
             Err(CliError::Flag(err)) => err.exit(),
             Err(CliError::Csv(err)) => {
-                eprintln!("{}", err);
+                eprintln!(
+                    "xan {}: {}",
+                    env::args().nth(1).unwrap_or("".to_string()),
+                    err
+                );
                 process::exit(1);
             }
             Err(CliError::Io(ref err)) if err.kind() == io::ErrorKind::BrokenPipe => {
                 process::exit(0);
             }
             Err(CliError::Io(err)) => {
-                eprintln!("{}", err);
+                eprintln!(
+                    "xan {}: {}",
+                    env::args().nth(1).unwrap_or("".to_string()),
+                    err
+                );
                 process::exit(1);
             }
             Err(CliError::Other(msg)) => {
-                eprintln!("{}", msg);
+                eprintln!(
+                    "xan {}: {}",
+                    env::args().nth(1).unwrap_or("".to_string()),
+                    msg
+                );
                 process::exit(1);
             }
-            // NOTE: I am not super fan to handle this as an error case...
-            Err(CliError::Help(usage)) => {
+            Err(CliError::Help(usage, exit_code)) => {
                 println!("{}", usage);
-                process::exit(0);
+                process::exit(exit_code);
             }
         },
     }
@@ -229,7 +257,6 @@ enum Command {
     Enum,
     Eval,
     Explode,
-    ForEach,
     F,
     Fill,
     Filter,
@@ -240,15 +267,17 @@ enum Command {
     Freq,
     Frequency,
     From,
+    #[serde(rename = "fuzzy-join")]
+    FuzzyJoin,
     Groupby,
     Guillotine,
     H,
+    Head,
     Headers,
     Heatmap,
     Help,
     Hist,
     Implode,
-    Index,
     Input,
     Join,
     Map,
@@ -258,11 +287,10 @@ enum Command {
     P,
     Parallel,
     Partition,
+    Pivot,
     Plot,
     Progress,
     Range,
-    #[serde(rename = "regex-join")]
-    RegexJoin,
     Rename,
     Reverse,
     Sample,
@@ -274,21 +302,22 @@ enum Command {
     Sort,
     Split,
     Stats,
+    Tail,
     To,
     Tokenize,
     Top,
     Transform,
     Transpose,
     T,
-    #[serde(rename = "url-join")]
-    UrlJoin,
+    Unpivot,
     V,
     View,
     Vocab,
+    Window,
 }
 
 impl Command {
-    fn run(self) -> CliResult<()> {
+    fn run(&self) -> CliResult<()> {
         let argv: Vec<_> = env::args().collect();
         let argv: Vec<_> = argv.iter().map(|s| &**s).collect();
         let argv = &*argv;
@@ -317,10 +346,11 @@ impl Command {
             Command::Flatmap => cmd::flatmap::run(argv),
             Command::Flatten | Command::F => cmd::flatten::run(argv),
             Command::Fmt => cmd::fmt::run(argv),
-            Command::ForEach => cmd::foreach::run(argv),
             Command::Freq | Command::Frequency => cmd::frequency::run(argv),
             Command::From => cmd::from::run(argv),
+            Command::FuzzyJoin => cmd::fuzzy_join::run(argv),
             Command::Groupby => cmd::groupby::run(argv),
+            Command::Head => cmd::head::run(argv),
             Command::Headers | Command::H => cmd::headers::run(argv),
             Command::Heatmap => cmd::heatmap::run(argv),
             Command::Help => {
@@ -333,7 +363,6 @@ impl Command {
             }
             Command::Hist => cmd::hist::run(argv),
             Command::Implode => cmd::implode::run(argv),
-            Command::Index => cmd::index::run(argv),
             Command::Input => cmd::input::run(argv),
             Command::Join => cmd::join::run(argv),
             Command::Network => cmd::network::run(argv),
@@ -342,10 +371,10 @@ impl Command {
             Command::Merge => cmd::merge::run(argv),
             Command::Parallel | Command::P => cmd::parallel::run(argv),
             Command::Partition => cmd::partition::run(argv),
+            Command::Pivot => cmd::pivot::run(argv),
             Command::Plot => cmd::plot::run(argv),
             Command::Progress => cmd::progress::run(argv),
             Command::Range => cmd::range::run(argv),
-            Command::RegexJoin => cmd::regex_join::run(argv),
             Command::Rename => cmd::rename::run(argv),
             Command::Reverse => cmd::reverse::run(argv),
             Command::Sample => cmd::sample::run(argv),
@@ -357,14 +386,16 @@ impl Command {
             Command::Sort => cmd::sort::run(argv),
             Command::Split => cmd::split::run(argv),
             Command::Stats => cmd::stats::run(argv),
+            Command::Tail => cmd::tail::run(argv),
             Command::To => cmd::to::run(argv),
             Command::Tokenize => cmd::tokenize::run(argv),
             Command::Top => cmd::top::run(argv),
             Command::Transform => cmd::transform::run(argv),
             Command::Transpose | Command::T => cmd::transpose::run(argv),
-            Command::UrlJoin => cmd::url_join::run(argv),
+            Command::Unpivot => cmd::unpivot::run(argv),
             Command::View | Command::V => cmd::view::run(argv),
             Command::Vocab => cmd::vocab::run(argv),
+            Command::Window => cmd::window::run(argv),
         }
     }
 }
@@ -377,7 +408,7 @@ pub enum CliError {
     Csv(csv::Error),
     Io(io::Error),
     Other(String),
-    Help(String),
+    Help(String, i32),
 }
 
 impl fmt::Display for CliError {
@@ -387,7 +418,7 @@ impl fmt::Display for CliError {
             CliError::Csv(ref e) => e.fmt(f),
             CliError::Io(ref e) => e.fmt(f),
             CliError::Other(ref s) => f.write_str(s),
-            CliError::Help(ref s) => f.write_str(s),
+            CliError::Help(ref s, _) => f.write_str(s),
         }
     }
 }
@@ -401,13 +432,16 @@ impl From<docopt::Error> for CliError {
                 let usage = util::colorize_help(&usage);
 
                 match kind.as_ref() {
-                    docopt::Error::Help => CliError::Help(usage),
-                    _ => CliError::Other(format!(
-                        "{}\n\n{} Use the {} flag for more information.",
-                        util::colorize_help(&usage),
-                        "Invalid command!".red(),
-                        "-h,--help".cyan()
-                    )),
+                    docopt::Error::Help => CliError::Help(usage, 0),
+                    _ => CliError::Help(
+                        format!(
+                            "{}\n\n{} Use the {} flag for more information.",
+                            util::colorize_help(&usage),
+                            "Invalid command!".red(),
+                            "-h,--help".cyan()
+                        ),
+                        1,
+                    ),
                 }
             }
             _ => CliError::Flag(err),
@@ -532,6 +566,15 @@ impl From<url::ParseError> for CliError {
     }
 }
 
+impl From<bgzip::BGZFError> for CliError {
+    fn from(value: bgzip::BGZFError) -> Self {
+        match value {
+            bgzip::BGZFError::IoError(err) => From::from(err),
+            _ => CliError::Other(value.to_string()),
+        }
+    }
+}
+
 impl From<()> for CliError {
     fn from(_: ()) -> CliError {
         CliError::Other("unknown error".to_string())
