@@ -1,17 +1,17 @@
 use std::borrow::Cow;
 use std::cmp::{max, Ordering, PartialOrd};
-use std::collections::HashMap;
 use std::fs::{self, File};
 use std::io::Read;
 use std::ops::{Add, Div, Mul, Neg, Rem, Sub};
 use std::path::{Path, PathBuf};
+use std::process::Command;
 use std::sync::Arc;
 
 use base64::prelude::*;
 use bstr::ByteSlice;
 use bytesize::ByteSize;
 use encoding::{label::encoding_from_whatwg_label, DecoderTrap};
-use flate2::read::GzDecoder;
+use flate2::read::MultiGzDecoder;
 use jiff::{fmt::strtime, tz::TimeZone, Timestamp, Zoned};
 use lazy_static::lazy_static;
 use mime2ext::mime2ext;
@@ -21,9 +21,11 @@ use paltoquet::{
     tokenizers::FingerprintTokenizer,
 };
 use rand::Rng;
+use regex::Regex;
 use unidecode::unidecode;
 use uuid::Uuid;
 
+use crate::collections::HashMap;
 use crate::dates;
 use crate::urls::LRUStems;
 
@@ -69,7 +71,6 @@ pub fn get_function(name: &str) -> Option<(Function, FunctionArguments)> {
             |args| variadic_arithmetic_op(args, Add::add),
             FunctionArguments::variadic(2),
         ),
-        "and" => (and, FunctionArguments::variadic(2)),
         "argmax" => (
             |args| argcompare(args, Ordering::is_gt),
             FunctionArguments::with_range(1..=2),
@@ -84,7 +85,7 @@ pub fn get_function(name: &str) -> Option<(Function, FunctionArguments)> {
             |args| unary_arithmetic_op(args, DynamicNumber::ceil),
             FunctionArguments::unary(),
         ),
-        "coalesce" => (coalesce, FunctionArguments::variadic(2)),
+        "cmd" => (cmd, FunctionArguments::binary()),
         "compact" => (compact, FunctionArguments::unary()),
         "concat" => (concat, FunctionArguments::variadic(2)),
         "contains" => (contains, FunctionArguments::binary()),
@@ -130,12 +131,23 @@ pub fn get_function(name: &str) -> Option<(Function, FunctionArguments)> {
         "last" => (last, FunctionArguments::unary()),
         "len" => (len, FunctionArguments::unary()),
         "log" => (
-            |args| unary_arithmetic_op(args, DynamicNumber::ln),
+            |args| match args.len() {
+                1 => unary_arithmetic_op(args, DynamicNumber::ln),
+                2 => binary_arithmetic_op(args, DynamicNumber::log),
+                _ => unreachable!(),
+            },
+            FunctionArguments::with_range(1..=2),
+        ),
+        "log2" => (
+            |args| unary_arithmetic_op(args, DynamicNumber::log2),
+            FunctionArguments::unary(),
+        ),
+        "log10" => (
+            |args| unary_arithmetic_op(args, DynamicNumber::log10),
             FunctionArguments::unary(),
         ),
         "lower" => (lower, FunctionArguments::unary()),
         "lru" => (lru, FunctionArguments::unary()),
-        "ltrim" => (ltrim, FunctionArguments::with_range(1..=2)),
         "match" => (regex_match, FunctionArguments::with_range(2..=3)),
         "max" => (variadic_max, FunctionArguments::variadic(2)),
         "md5" => (md5, FunctionArguments::unary()),
@@ -164,7 +176,18 @@ pub fn get_function(name: &str) -> Option<(Function, FunctionArguments)> {
             FunctionArguments::unary(),
         ),
         "not" => (not, FunctionArguments::unary()),
-        "or" => (or, FunctionArguments::variadic(2)),
+        "pad" => (
+            |args| pad(pad::Alignment::Middle, args),
+            FunctionArguments::with_range(2..=3),
+        ),
+        "lpad" => (
+            |args| pad(pad::Alignment::Right, args),
+            FunctionArguments::with_range(2..=3),
+        ),
+        "rpad" => (
+            |args| pad(pad::Alignment::Left, args),
+            FunctionArguments::with_range(2..=3),
+        ),
         "parse_dataurl" => (parse_dataurl, FunctionArguments::unary()),
         "parse_json" => (parse_json, FunctionArguments::unary()),
         "pjoin" | "pathjoin" => (pathjoin, FunctionArguments::variadic(2)),
@@ -172,6 +195,7 @@ pub fn get_function(name: &str) -> Option<(Function, FunctionArguments)> {
             |args| binary_arithmetic_op(args, DynamicNumber::pow),
             FunctionArguments::binary(),
         ),
+        "printf" => (printf, FunctionArguments::variadic(2)),
         "random" => (random, FunctionArguments::nullary()),
         "read" => (
             read,
@@ -183,12 +207,14 @@ pub fn get_function(name: &str) -> Option<(Function, FunctionArguments)> {
         ),
         "read_csv" => (read_csv, FunctionArguments::unary()),
         "read_json" => (read_json, FunctionArguments::unary()),
+        "regex" => (parse_regex, FunctionArguments::unary()),
         "replace" => (replace, FunctionArguments::nary(3)),
         "round" => (
             |args| unary_arithmetic_op(args, DynamicNumber::round),
             FunctionArguments::unary(),
         ),
-        "rtrim" => (rtrim, FunctionArguments::with_range(1..=2)),
+        "shell" => (shell, FunctionArguments::unary()),
+        "shlex_split" => (shlex_split, FunctionArguments::unary()),
         "slice" => (slice, FunctionArguments::with_range(2..=3)),
         "split" => (split, FunctionArguments::with_range(2..=3)),
         "sqrt" => (
@@ -236,9 +262,12 @@ pub fn get_function(name: &str) -> Option<(Function, FunctionArguments)> {
         ),
         "timestamp" => (timestamp, FunctionArguments::unary()),
         "timestamp_ms" => (timestamp_ms, FunctionArguments::unary()),
+        "to_fixed" => (to_fixed, FunctionArguments::binary()),
         "to_timezone" => (to_timezone, FunctionArguments::nary(3)),
         "to_local_timezone" => (to_local_timezone, FunctionArguments::binary()),
         "trim" => (trim, FunctionArguments::with_range(1..=2)),
+        "ltrim" => (ltrim, FunctionArguments::with_range(1..=2)),
+        "rtrim" => (rtrim, FunctionArguments::with_range(1..=2)),
         "trunc" => (
             |args| unary_arithmetic_op(args, DynamicNumber::trunc),
             FunctionArguments::unary(),
@@ -267,52 +296,64 @@ pub fn get_function(name: &str) -> Option<(Function, FunctionArguments)> {
 }
 
 // Strings
-fn trim(args: BoundArguments) -> FunctionResult {
-    let chars_opt = args.get(1);
+macro_rules! make_trim_fn {
+    ($name: ident, $trim: ident, $trim_matches: ident) => {
+        fn $name(args: BoundArguments) -> FunctionResult {
+            let chars_opt = args.get(1);
 
-    Ok(match chars_opt {
-        None => match args.get1() {
-            DynamicValue::Bytes(bytes) => DynamicValue::from(bytes.trim()),
-            value => DynamicValue::from(value.try_as_str()?.trim()),
-        },
-        Some(chars) => {
-            let pattern = chars.try_as_str()?.chars().collect::<Vec<char>>();
-            DynamicValue::from(args.get1_str()?.trim_matches(|c| pattern.contains(&c)))
+            Ok(match chars_opt {
+                None => match args.get1() {
+                    DynamicValue::Bytes(bytes) => DynamicValue::from(bytes.$trim()),
+                    value => DynamicValue::from(value.try_as_str()?.$trim()),
+                },
+                Some(chars) => {
+                    let pattern = chars.try_as_str()?.chars().collect::<Vec<char>>();
+                    DynamicValue::from(args.get1_str()?.$trim_matches(|c| pattern.contains(&c)))
+                }
+            })
         }
-    })
+    };
 }
 
-fn ltrim(args: BoundArguments) -> FunctionResult {
-    let chars_opt = args.get(1);
+make_trim_fn!(trim, trim, trim_matches);
+make_trim_fn!(ltrim, trim_start, trim_start_matches);
+make_trim_fn!(rtrim, trim_end, trim_end_matches);
 
-    Ok(match chars_opt {
-        None => match args.get1() {
-            DynamicValue::Bytes(bytes) => DynamicValue::from(bytes.trim_start()),
-            value => DynamicValue::from(value.try_as_str()?.trim_start()),
-        },
-        Some(chars) => {
-            let pattern = chars.try_as_str()?.chars().collect::<Vec<char>>();
-            DynamicValue::from(
-                args.get1_str()?
-                    .trim_start_matches(|c| pattern.contains(&c)),
-            )
+fn pad(alignment: pad::Alignment, args: BoundArguments) -> FunctionResult {
+    use pad::PadStr;
+
+    let mut args_iter = args.into_iter();
+    let first_arg = args_iter.next().unwrap();
+    let string = first_arg.try_as_str()?;
+
+    let width = args_iter.next().unwrap().try_as_usize()?;
+    let padding_char = match args_iter.next() {
+        None => ' ',
+        Some(value) => {
+            let padding_string = value.try_as_str()?;
+
+            match padding_string.chars().count() {
+                0 => {
+                    return Err(EvaluationError::Custom(
+                        "provided padding char is empty".to_string(),
+                    ));
+                }
+                1 => padding_string.chars().next().unwrap(),
+                2.. => {
+                    return Err(EvaluationError::Custom(
+                        "provided padding char is longer than a char".to_string(),
+                    ));
+                }
+            }
         }
-    })
-}
+    };
 
-fn rtrim(args: BoundArguments) -> FunctionResult {
-    let chars_opt = args.get(1);
-
-    Ok(match chars_opt {
-        None => match args.get1() {
-            DynamicValue::Bytes(bytes) => DynamicValue::from(bytes.trim_end()),
-            value => DynamicValue::from(value.try_as_str()?.trim_end()),
-        },
-        Some(chars) => {
-            let pattern = chars.try_as_str()?.chars().collect::<Vec<char>>();
-            DynamicValue::from(args.get1_str()?.trim_end_matches(|c| pattern.contains(&c)))
-        }
-    })
+    Ok(DynamicValue::from(string.pad(
+        width,
+        padding_char,
+        alignment,
+        false,
+    )))
 }
 
 fn escape_regex(args: BoundArguments) -> FunctionResult {
@@ -328,16 +369,29 @@ fn md5(args: BoundArguments) -> FunctionResult {
 
 fn split(args: BoundArguments) -> FunctionResult {
     let to_split = args.get(0).unwrap().try_as_str()?;
-    let pattern = args.get(1).unwrap().try_as_str()?;
+    let pattern_arg = args.get(1).unwrap();
     let count = args.get(2);
 
-    let splitted: Vec<DynamicValue> = if let Some(c) = count {
-        to_split
-            .splitn(c.try_as_usize()? + 1, pattern.as_ref())
-            .map(DynamicValue::from)
-            .collect()
+    let splitted: Vec<DynamicValue> = if let DynamicValue::Regex(pattern) = pattern_arg {
+        if let Some(c) = count {
+            pattern
+                .splitn(&to_split, c.try_as_usize()? + 1)
+                .map(DynamicValue::from)
+                .collect()
+        } else {
+            pattern.split(&to_split).map(DynamicValue::from).collect()
+        }
     } else {
-        to_split.split(&*pattern).map(DynamicValue::from).collect()
+        let pattern = pattern_arg.try_as_str()?;
+
+        if let Some(c) = count {
+            to_split
+                .splitn(c.try_as_usize()? + 1, pattern.as_ref())
+                .map(DynamicValue::from)
+                .collect()
+        } else {
+            to_split.split(&*pattern).map(DynamicValue::from).collect()
+        }
     };
 
     Ok(DynamicValue::from(splitted))
@@ -508,6 +562,44 @@ fn fmt_number(mut args: BoundArguments) -> FunctionResult {
     Ok(DynamicValue::from(crate::util::format_number(number)))
 }
 
+fn printf(args: BoundArguments) -> FunctionResult {
+    let l = args.len() - 1;
+
+    let mut args_iter = args.into_iter();
+    let fmt_arg = args_iter.next().unwrap();
+    let fmt = fmt_arg.try_as_str()?;
+
+    let mut fmt_args: Vec<Box<dyn sprintf::Printf>> = Vec::with_capacity(l);
+
+    for arg in args_iter {
+        match arg {
+            DynamicValue::Integer(i) => fmt_args.push(Box::new(i)),
+            DynamicValue::Float(f) => fmt_args.push(Box::new(f)),
+            _ => fmt_args.push(Box::new(arg.try_as_str()?.into_owned())),
+        };
+    }
+
+    let fmt_args_refs = fmt_args.iter().map(|b| b.as_ref()).collect::<Vec<_>>();
+
+    match sprintf::vsprintf(&fmt, &fmt_args_refs) {
+        Ok(string) => Ok(DynamicValue::from(string)),
+        Err(_) => Err(EvaluationError::Custom(
+            "printf formatting error".to_string(),
+        )),
+    }
+}
+
+fn to_fixed(mut args: BoundArguments) -> FunctionResult {
+    let (arg1, arg2) = args.pop2();
+
+    let n = arg1.try_as_f64()?;
+    let p = arg2.try_as_usize()?.min(16);
+
+    let formatted = format!("{:.precision$}", n, precision = p);
+
+    Ok(DynamicValue::from(formatted))
+}
+
 // Lists & Sequences
 fn first(mut args: BoundArguments) -> FunctionResult {
     let arg = args.pop1();
@@ -565,6 +657,22 @@ fn get_subroutine<'a>(
                 value.chars().nth(index as usize).map(DynamicValue::from)
             }
         }
+        DynamicValue::Bytes(value) => {
+            let mut index = key.try_as_i64()?;
+
+            if index < 0 {
+                index += value.len() as i64;
+            }
+
+            if index < 0 {
+                None
+            } else {
+                let value =
+                    std::str::from_utf8(value).map_err(|_| EvaluationError::UnicodeDecodeError)?;
+
+                value.chars().nth(index as usize).map(DynamicValue::from)
+            }
+        }
         DynamicValue::List(list) => {
             let mut index = key.try_as_i64()?;
 
@@ -631,6 +739,8 @@ fn slice(args: BoundArguments) -> FunctionResult {
                     lo = max(0, l as i64 + lo);
 
                     list[..lo as usize].to_vec()
+                } else if lo >= list.len() as i64 {
+                    Vec::new()
                 } else {
                     list[..lo as usize].to_vec()
                 }
@@ -638,15 +748,33 @@ fn slice(args: BoundArguments) -> FunctionResult {
             Some(hi_value) => {
                 let mut hi = hi_value.try_as_i64()?;
 
-                if lo < 0 {
+                if lo >= list.len() as i64 {
                     Vec::new()
+                } else if lo < 0 {
+                    let l = list.len();
+
+                    lo = max(0, l as i64 + lo);
+
+                    if hi < 0 {
+                        hi = max(0, l as i64 + hi);
+                    }
+
+                    if hi <= lo {
+                        Vec::new()
+                    } else {
+                        list[lo as usize..hi.min(list.len() as i64) as usize].to_vec()
+                    }
                 } else {
                     if hi < 0 {
                         let l = list.len();
                         hi = max(0, l as i64 + hi);
                     }
 
-                    list[lo as usize..(hi - lo) as usize].to_vec()
+                    if hi <= lo {
+                        Vec::new()
+                    } else {
+                        list[lo as usize..hi.min(list.len() as i64) as usize].to_vec()
+                    }
                 }
             }
         };
@@ -676,7 +804,18 @@ fn slice(args: BoundArguments) -> FunctionResult {
             let mut hi = hi_value.try_as_i64()?;
 
             if lo < 0 {
-                "".to_string()
+                let l = string.chars().count();
+                lo = max(0, l as i64 + lo);
+
+                if hi < 0 {
+                    hi = max(0, l as i64 + hi);
+                }
+
+                if hi <= lo {
+                    "".to_string()
+                } else {
+                    chars.skip(lo as usize).take((hi - lo) as usize).collect()
+                }
             } else {
                 if hi < 0 {
                     let l = string.chars().count();
@@ -997,59 +1136,11 @@ where
     Ok(DynamicValue::from(min_item.map(|t| t.1)))
 }
 
-// Utilities
-fn coalesce(args: BoundArguments) -> FunctionResult {
-    for arg in args {
-        if arg.is_truthy() {
-            return Ok(arg);
-        }
-    }
-
-    Ok(DynamicValue::None)
-}
-
 // Boolean
 fn not(mut args: BoundArguments) -> FunctionResult {
     Ok(DynamicValue::from(!args.pop1_bool()))
 }
 
-fn and(args: BoundArguments) -> FunctionResult {
-    let mut last: Option<DynamicValue> = None;
-
-    for arg in args {
-        if arg.is_falsey() {
-            return Ok(arg);
-        }
-
-        last = Some(arg);
-    }
-
-    Ok(last.unwrap())
-}
-
-fn or(args: BoundArguments) -> FunctionResult {
-    let mut last: Option<DynamicValue> = None;
-
-    for arg in args {
-        if arg.is_truthy() {
-            return Ok(arg);
-        }
-
-        last = Some(arg);
-    }
-
-    Ok(last.unwrap())
-}
-
-// TODO: rewrap those to take lists instead, since the variadic usage is mostly moot
-// fn all(args: BoundArguments) -> FunctionResult {
-//     Ok(DynamicValue::from(args.into_iter().all(|v| v.is_truthy())))
-// }
-
-// fn any(args: BoundArguments) -> FunctionResult {
-//     Ok(DynamicValue::from(args.into_iter().any(|v| v.is_truthy())))
-// }
-
 // Comparison
 fn abstract_compare<F>(mut args: BoundArguments, validate: F) -> FunctionResult
 where
@@ -1182,7 +1273,7 @@ fn abstract_read(
             let mut buffer: Vec<u8> = Vec::new();
 
             if path.ends_with(".gz") {
-                let mut gz = GzDecoder::new(file);
+                let mut gz = MultiGzDecoder::new(file);
                 gz.read_to_end(&mut buffer)
                     .map_err(|_| EvaluationError::IO(format!("cannot read file {}", path)))?;
             } else {
@@ -1198,7 +1289,7 @@ fn abstract_read(
             let mut buffer = String::new();
 
             if path.ends_with(".gz") {
-                let mut gz = GzDecoder::new(file);
+                let mut gz = MultiGzDecoder::new(file);
                 gz.read_to_string(&mut buffer)
                     .map_err(|_| EvaluationError::IO(format!("cannot read file {}", path)))?;
             } else {
@@ -1223,7 +1314,8 @@ fn read(args: BoundArguments) -> FunctionResult {
 
 fn read_json(args: BoundArguments) -> FunctionResult {
     let contents = abstract_read(args.get1(), None, None)?;
-    serde_json::from_str(&contents).map_err(|_| EvaluationError::JSONParseError)
+    serde_json::from_str(&contents)
+        .map_err(|_| EvaluationError::JSONParseError(format!("{:?}", contents)))
 }
 
 fn read_csv(args: BoundArguments) -> FunctionResult {
@@ -1546,7 +1638,8 @@ fn err(args: BoundArguments) -> FunctionResult {
 fn parse_json(args: BoundArguments) -> FunctionResult {
     let arg = args.get1_str()?;
 
-    serde_json::from_str(arg.as_ref()).map_err(|_| EvaluationError::JSONParseError)
+    serde_json::from_str(arg.as_ref())
+        .map_err(|_| EvaluationError::JSONParseError(format!("{:?}", args.get1())))
 }
 
 fn parse_dataurl(args: BoundArguments) -> FunctionResult {
@@ -1597,3 +1690,93 @@ fn html_unescape(args: BoundArguments) -> FunctionResult {
         &string,
     )))
 }
+
+fn parse_regex(args: BoundArguments) -> FunctionResult {
+    let string = args.get1_str()?;
+
+    Ok(DynamicValue::from(Regex::new(&string).map_err(|_| {
+        EvaluationError::Custom(format!("could not parse \"{}\" as regex", string))
+    })?))
+}
+
+fn shlex_split(args: BoundArguments) -> FunctionResult {
+    let string = args.get1_str()?;
+
+    if let Some(splitted) = shlex::split(&string) {
+        Ok(DynamicValue::from(
+            splitted
+                .into_iter()
+                .map(DynamicValue::from)
+                .collect::<Vec<_>>(),
+        ))
+    } else {
+        Err(EvaluationError::Custom(format!(
+            "could not split {:?}",
+            args.get1()
+        )))
+    }
+}
+
+fn cmd(mut args: BoundArguments) -> FunctionResult {
+    let (command_name_arg, command_args) = args.pop2();
+
+    let command_name = command_name_arg.try_as_str()?;
+
+    let mut command = Command::new(command_name.as_ref());
+
+    for command_arg in command_args.try_as_list()? {
+        command.arg(command_arg.try_as_str()?.as_ref());
+    }
+
+    if let Ok(mut output) = command.output() {
+        if output.status.success() {
+            let result = &mut output.stdout;
+            result.truncate(result.trim_ascii_end().len());
+
+            Ok(DynamicValue::from_owned_bytes(output.stdout))
+        } else {
+            Err(EvaluationError::Custom(format!(
+                "\"{}\" failed!",
+                command_name
+            )))
+        }
+    } else {
+        Err(EvaluationError::Custom(format!(
+            "error while spawning \"{}\"",
+            command_name
+        )))
+    }
+}
+
+fn shell(args: BoundArguments) -> FunctionResult {
+    let pipeline = args.get1_str()?;
+
+    let mut command = if cfg!(target_os = "windows") {
+        let mut command = Command::new("cmd");
+        command.args(["/C", pipeline.as_ref()]);
+        command
+    } else {
+        let mut command = Command::new(std::env::var("SHELL").unwrap_or_else(|_| "sh".to_string()));
+        command.args(["-c", pipeline.as_ref()]);
+        command
+    };
+
+    if let Ok(mut output) = command.output() {
+        if output.status.success() {
+            let result = &mut output.stdout;
+            result.truncate(result.trim_ascii_end().len());
+
+            Ok(DynamicValue::from_owned_bytes(output.stdout))
+        } else {
+            Err(EvaluationError::Custom(format!(
+                "shell pipeline \"{}\" failed!",
+                pipeline
+            )))
+        }
+    } else {
+        Err(EvaluationError::Custom(format!(
+            "error while running shell pipeline \"{}\"",
+            pipeline
+        )))
+    }
+}
