@@ -1,10 +1,11 @@
-use std::collections::HashMap;
 use std::fmt;
 
 use arrayvec::ArrayVec;
 use csv::ByteRecord;
 use regex::RegexBuilder;
 
+use crate::collections::HashMap;
+
 use super::error::{ConcretizationError, EvaluationError, SpecifiedEvaluationError};
 use super::functions::{get_function, Function};
 use super::parser::{parse_expression, Expr, FunctionCall};
@@ -14,11 +15,6 @@ use super::types::{
     HeadersIndex, LambdaArguments, BOUND_ARGUMENTS_CAPACITY,
 };
 
-// NOTE: the evaluation arguments could be splitted into transient evaluation
-// context that MUST be mutated before each evaluation and needs to be managed
-// by the user or a threadsafe program, and permanent evaluation context, such
-// as the headers index, computed at concretization time.
-
 #[derive(Debug, Default, Clone)]
 pub struct GlobalVariables {
     slots: ArrayVec<(String, DynamicValue), 2>,
@@ -58,20 +54,75 @@ impl GlobalVariables {
     }
 }
 
-#[derive(Debug, Clone, Default)]
-pub struct EvaluationContext {
-    headers_index: HeadersIndex,
+#[derive(Debug, Clone)]
+pub struct EvaluationContext<'a> {
+    pub index: Option<usize>,
+    pub record: &'a ByteRecord,
+    pub headers_index: &'a HeadersIndex,
+    pub globals: Option<&'a GlobalVariables>,
+    pub lambda_variables: Option<&'a LambdaArguments>,
+    pub last_value: Option<DynamicValue>,
 }
 
-impl EvaluationContext {
-    pub fn new<'a>(headers: impl IntoIterator<Item = &'a [u8]>) -> Self {
+impl<'a> EvaluationContext<'a> {
+    pub fn new(
+        index: Option<usize>,
+        record: &'a ByteRecord,
+        headers_index: &'a HeadersIndex,
+    ) -> Self {
         Self {
-            headers_index: HeadersIndex::from_headers(headers),
+            index,
+            record,
+            headers_index,
+            globals: None,
+            lambda_variables: None,
+            last_value: None,
         }
     }
 
-    pub fn get_column_index(&self, indexation: &ColumIndexationBy) -> Option<usize> {
-        self.headers_index.get(indexation)
+    pub fn new_with_globals(
+        index: Option<usize>,
+        record: &'a ByteRecord,
+        headers_index: &'a HeadersIndex,
+        globals: &'a GlobalVariables,
+    ) -> Self {
+        let mut context = Self::new(index, record, headers_index);
+        context.globals = Some(globals);
+
+        context
+    }
+
+    fn dummy(record: &'a ByteRecord, headers_index: &'a HeadersIndex) -> Self {
+        Self {
+            index: None,
+            record,
+            headers_index,
+            globals: None,
+            lambda_variables: None,
+            last_value: None,
+        }
+    }
+
+    pub fn with_lambda_variables(&self, variables: &'a LambdaArguments) -> Self {
+        Self {
+            index: self.index,
+            record: self.record,
+            headers_index: self.headers_index,
+            globals: self.globals,
+            lambda_variables: Some(variables),
+            last_value: self.last_value.clone(),
+        }
+    }
+
+    pub fn with_globals(&self, globals: &'a GlobalVariables) -> Self {
+        Self {
+            index: self.index,
+            record: self.record,
+            headers_index: self.headers_index,
+            globals: Some(globals),
+            lambda_variables: self.lambda_variables,
+            last_value: self.last_value.clone(),
+        }
     }
 }
 
@@ -86,6 +137,8 @@ pub enum ConcreteExpr {
     Map(Vec<(String, ConcreteExpr)>),
     Call(ConcreteFunctionCall),
     SpecialCall(ConcreteSpecialFunctionCall),
+    Pipeline(Vec<ConcreteExpr>),
+    Underscore,
 }
 
 // NOTE: the bind/evaluate distinction is still useful to propagate the calling
@@ -95,6 +148,43 @@ impl ConcreteExpr {
         matches!(self, Self::Value(_))
     }
 
+    pub fn used_column_indices(&self, scratch: &mut Vec<usize>) {
+        match self {
+            Self::Column(i) => {
+                scratch.push(*i);
+            }
+            Self::Lambda(_, inner) => {
+                inner.used_column_indices(scratch);
+            }
+            Self::List(values) => {
+                for value in values {
+                    value.used_column_indices(scratch);
+                }
+            }
+            Self::Map(map) => {
+                for (_, value) in map {
+                    value.used_column_indices(scratch);
+                }
+            }
+            Self::Call(call) => {
+                for arg in call.args.iter() {
+                    arg.used_column_indices(scratch);
+                }
+            }
+            Self::SpecialCall(call) => {
+                for arg in call.args.iter() {
+                    arg.used_column_indices(scratch);
+                }
+            }
+            Self::Pipeline(exprs) => {
+                for expr in exprs {
+                    expr.used_column_indices(scratch);
+                }
+            }
+            _ => (),
+        };
+    }
+
     // NOTE: here we are not abiding by the DFS
     fn is_deeply_statically_evaluable(&self, bound: &Vec<String>) -> bool {
         match self {
@@ -145,56 +235,46 @@ impl ConcreteExpr {
         }
     }
 
-    fn bind(
-        &self,
-        record: &ByteRecord,
-        globals: Option<&GlobalVariables>,
-        lambda_variables: Option<&LambdaArguments>,
-    ) -> Result<DynamicValue, EvaluationError> {
+    fn bind(&self, context: &EvaluationContext) -> Result<DynamicValue, EvaluationError> {
         Ok(match self {
             Self::Value(value) => value.clone(),
-            Self::Column(index) => match record.get(*index) {
+            Self::Column(index) => match context.record.get(*index) {
                 None => return Err(EvaluationError::ColumnOutOfRange(*index)),
                 Some(cell) => DynamicValue::from(cell),
             },
             Self::GlobalVariable(index) => {
-                match globals.expect("globals were not set!").get(*index) {
+                match context.globals.expect("globals were not set!").get(*index) {
                     None => return Err(EvaluationError::GlobalVariableOutOfRange(*index)),
                     Some(value) => value.clone(),
                 }
             }
-            Self::LambdaBinding(name) => lambda_variables
+            Self::LambdaBinding(name) => context
+                .lambda_variables
                 .expect("lambda_variables MUST be set")
                 .get(name)
                 .clone(),
+            Self::Underscore => match context.last_value.as_ref() {
+                None => return Err(EvaluationError::UnfillableUnderscore),
+                Some(last_value) => last_value.clone(),
+            },
             Self::List(_)
             | Self::Map(_)
             | Self::Call(_)
             | Self::SpecialCall(_)
-            | Self::Lambda(_, _) => unreachable!(),
+            | Self::Lambda(_, _)
+            | Self::Pipeline(_) => unreachable!(),
         })
     }
 
-    pub fn evaluate(
-        &self,
-        index: Option<usize>,
-        record: &ByteRecord,
-        context: &EvaluationContext,
-        globals: Option<&GlobalVariables>,
-        lambda_variables: Option<&LambdaArguments>,
-    ) -> EvaluationResult {
+    pub fn evaluate(&self, context: &EvaluationContext) -> EvaluationResult {
         match self {
-            Self::Call(function_call) => {
-                function_call.run(index, record, context, globals, lambda_variables)
-            }
-            Self::SpecialCall(function_call) => {
-                function_call.run(index, record, context, globals, lambda_variables)
-            }
+            Self::Call(function_call) => function_call.run(context),
+            Self::SpecialCall(function_call) => function_call.run(context),
             Self::List(items) => {
                 let mut bound = Vec::with_capacity(items.len());
 
                 for item in items {
-                    bound.push(item.evaluate(index, record, context, globals, lambda_variables)?);
+                    bound.push(item.evaluate(context)?);
                 }
 
                 Ok(DynamicValue::from(bound))
@@ -203,17 +283,24 @@ impl ConcreteExpr {
                 let mut bound = HashMap::with_capacity(pairs.len());
 
                 for (k, v) in pairs {
-                    bound.insert(
-                        k.to_string(),
-                        v.evaluate(index, record, context, globals, lambda_variables)?,
-                    );
+                    bound.insert(k.to_string(), v.evaluate(context)?);
                 }
 
                 Ok(DynamicValue::from(bound))
             }
-            _ => self
-                .bind(record, globals, lambda_variables)
-                .map_err(|err| err.anonymous()),
+            Self::Pipeline(pipeline) => {
+                debug_assert!(!pipeline.is_empty());
+
+                let mut pipeline_context = context.clone();
+
+                for expr in pipeline {
+                    let value = expr.evaluate(&pipeline_context)?;
+                    pipeline_context.last_value = Some(value);
+                }
+
+                Ok(pipeline_context.last_value.unwrap())
+            }
+            _ => self.bind(context).map_err(|err| err.anonymous()),
         }
     }
 }
@@ -238,56 +325,27 @@ impl ConcreteFunctionCall {
     }
 
     fn static_run(&self) -> EvaluationResult {
-        self.run(
-            None,
-            &ByteRecord::new(),
-            &EvaluationContext::default(),
-            None,
-            None,
-        )
+        let record = ByteRecord::new();
+        let headers_index = HeadersIndex::default();
+
+        self.run(&EvaluationContext::dummy(&record, &headers_index))
     }
 
-    fn run(
-        &self,
-        index: Option<usize>,
-        record: &ByteRecord,
-        context: &EvaluationContext,
-        globals: Option<&GlobalVariables>,
-        lambda_variables: Option<&LambdaArguments>,
-    ) -> EvaluationResult {
+    fn run(&self, context: &EvaluationContext) -> EvaluationResult {
         let mut bound_args = BoundArguments::new();
 
         for arg in self.args.iter() {
             match arg {
                 ConcreteExpr::Call(sub_function_call) => {
-                    bound_args.push(sub_function_call.run(
-                        index,
-                        record,
-                        context,
-                        globals,
-                        lambda_variables,
-                    )?);
+                    bound_args.push(sub_function_call.run(context)?);
                 }
                 ConcreteExpr::SpecialCall(sub_function_call) => {
-                    bound_args.push(sub_function_call.run(
-                        index,
-                        record,
-                        context,
-                        globals,
-                        lambda_variables,
-                    )?);
+                    bound_args.push(sub_function_call.run(context)?);
                 }
-                ConcreteExpr::List(_) | ConcreteExpr::Map(_) => bound_args.push(arg.evaluate(
-                    index,
-                    record,
-                    context,
-                    globals,
-                    lambda_variables,
-                )?),
-                _ => bound_args.push(
-                    arg.bind(record, globals, lambda_variables)
-                        .map_err(|err| err.specify(&self.name))?,
-                ),
+                ConcreteExpr::List(_) | ConcreteExpr::Map(_) => {
+                    bound_args.push(arg.evaluate(context)?)
+                }
+                _ => bound_args.push(arg.bind(context).map_err(|err| err.specify(&self.name))?),
             }
         }
 
@@ -321,7 +379,11 @@ impl ConcreteSpecialFunctionCall {
     fn is_statically_evaluable(&self, bound: &Vec<String>) -> bool {
         // NOTE: other special function are not suitable for late
         // statical evaluation.
-        if ["col", "cols", "headers", "index"].contains(&self.name.as_str()) {
+        if [
+            "col", "cols", "headers", "index", "if", "unless", "warn", "and", "or",
+        ]
+        .contains(&self.name.as_str())
+        {
             return false;
         }
 
@@ -331,31 +393,14 @@ impl ConcreteSpecialFunctionCall {
     }
 
     fn static_run(&self) -> EvaluationResult {
-        self.run(
-            None,
-            &ByteRecord::new(),
-            &EvaluationContext::default(),
-            None,
-            None,
-        )
+        let record = ByteRecord::new();
+        let headers_index = HeadersIndex::default();
+
+        self.run(&EvaluationContext::dummy(&record, &headers_index))
     }
 
-    fn run(
-        &self,
-        index: Option<usize>,
-        record: &ByteRecord,
-        context: &EvaluationContext,
-        globals: Option<&GlobalVariables>,
-        lambda_variables: Option<&LambdaArguments>,
-    ) -> EvaluationResult {
-        (self.function)(
-            index,
-            record,
-            context,
-            &self.args,
-            globals,
-            lambda_variables,
-        )
+    fn run(&self, context: &EvaluationContext) -> EvaluationResult {
+        (self.function)(context, &self.args)
     }
 }
 
@@ -421,10 +466,56 @@ fn concretize_call(
             }
         }
 
+        // TODO: at some point it might be nice to refactor comptime evaluation
+        // to better take care of complex cases such as branching & trying.
+
+        // NOTE: special cases wrt branching
+        if function_name == "if" || function_name == "unless" {
+            let condition = concretize_expression(call.args[0].clone().1, headers, globals)?;
+
+            if let Ok(value) = condition.try_unwrap() {
+                let path = if function_name == "if" {
+                    if value.is_truthy() {
+                        1
+                    } else {
+                        2
+                    }
+                } else if value.is_truthy() {
+                    2
+                } else {
+                    1
+                };
+
+                return concretize_expression(call.args[path].clone().1, headers, globals);
+            }
+        }
+
+        if function_name == "or" || function_name == "and" {
+            for (_, arg) in call.args.iter() {
+                if let Ok(condition) = concretize_expression(arg.clone(), headers, globals) {
+                    if let Ok(value) = condition.try_unwrap() {
+                        let test = if function_name == "or" {
+                            value.is_truthy()
+                        } else {
+                            value.is_falsey()
+                        };
+
+                        if test {
+                            return Ok(ConcreteExpr::Value(value));
+                        }
+                    } else {
+                        break;
+                    }
+                } else {
+                    break;
+                }
+            }
+        }
+
         let concrete_args = concretize_arguments(&arguments, call.args, headers, globals);
 
         // NOTE: special case of bubbling-up exceptions
-        if function_name.as_str() == "try" && concrete_args.is_err() {
+        if function_name == "try" && concrete_args.is_err() {
             return Ok(ConcreteExpr::Value(DynamicValue::None));
         }
 
@@ -527,9 +618,7 @@ pub fn concretize_expression(
     globals: Option<&GlobalVariables>,
 ) -> Result<ConcreteExpr, ConcretizationError> {
     Ok(match expr {
-        Expr::Underscore => {
-            panic!("found underscore while concretizing expression!\nThis suggests that moonblade::parser::Expr::fill_underscore is not doing its job properly.");
-        }
+        Expr::Underscore => ConcreteExpr::Underscore,
         Expr::Null => ConcreteExpr::Value(DynamicValue::None),
         Expr::Bool(v) => ConcreteExpr::Value(DynamicValue::Boolean(v)),
         Expr::Float(v) => ConcreteExpr::Value(DynamicValue::Float(v)),
@@ -545,7 +634,7 @@ pub fn concretize_expression(
 
             let indexation = ColumIndexationBy::Name(name);
 
-            match indexation.find_column_index(headers, headers.len()) {
+            match indexation.find_column_index(headers) {
                 Some(index) => ConcreteExpr::Column(index),
                 None => {
                     if unsure {
@@ -570,6 +659,12 @@ pub fn concretize_expression(
             names,
             Box::new(concretize_expression(*expr, headers, globals)?),
         ),
+        Expr::Pipeline(pipeline) => ConcreteExpr::Pipeline(
+            pipeline
+                .into_iter()
+                .map(|expr| concretize_expression(expr, headers, globals))
+                .collect::<Result<Vec<_>, _>>()?,
+        ),
         Expr::LambdaBinding(name) => ConcreteExpr::LambdaBinding(name),
         Expr::Slice(_) | Expr::StarSlice(_) => unreachable!(),
     })
@@ -579,37 +674,54 @@ pub fn eval_expression_with_globals(
     expr: &ConcreteExpr,
     index: Option<usize>,
     record: &ByteRecord,
-    context: &EvaluationContext,
+    headers_index: &HeadersIndex,
     globals: &GlobalVariables,
 ) -> Result<DynamicValue, SpecifiedEvaluationError> {
-    expr.evaluate(index, record, context, Some(globals), None)
+    let context = EvaluationContext::new_with_globals(index, record, headers_index, globals);
+
+    expr.evaluate(&context)
+}
+
+pub fn eval_expression_with_optional_last_value(
+    expr: &ConcreteExpr,
+    index: Option<usize>,
+    record: &ByteRecord,
+    headers_index: &HeadersIndex,
+    last_value: Option<DynamicValue>,
+) -> Result<DynamicValue, SpecifiedEvaluationError> {
+    let mut context = EvaluationContext::new(index, record, headers_index);
+    context.last_value = last_value;
+
+    expr.evaluate(&context)
 }
 
 pub fn eval_expression(
     expr: &ConcreteExpr,
     index: Option<usize>,
     record: &ByteRecord,
-    context: &EvaluationContext,
+    headers_index: &HeadersIndex,
 ) -> Result<DynamicValue, SpecifiedEvaluationError> {
-    expr.evaluate(index, record, context, None, None)
+    let context = EvaluationContext::new(index, record, headers_index);
+
+    expr.evaluate(&context)
 }
 
 #[derive(Clone, Debug)]
 pub struct Program {
     pub expr: ConcreteExpr,
-    context: EvaluationContext,
+    headers_index: HeadersIndex,
 }
 
 impl Program {
     pub fn parse(code: &str, headers: &ByteRecord) -> Result<Self, ConcretizationError> {
         let expr = match parse_expression(code) {
-            Err(_) => return Err(ConcretizationError::ParseError(code.to_string())),
+            Err(err) => return Err(ConcretizationError::ParseError(err)),
             Ok(parsed_expr) => concretize_expression(parsed_expr, headers, None)?,
         };
 
         Ok(Self {
             expr,
-            context: EvaluationContext::new(headers),
+            headers_index: HeadersIndex::from_headers(headers),
         })
     }
 
@@ -619,13 +731,13 @@ impl Program {
         globals: &GlobalVariables,
     ) -> Result<Self, ConcretizationError> {
         let expr = match parse_expression(code) {
-            Err(_) => return Err(ConcretizationError::ParseError(code.to_string())),
+            Err(err) => return Err(ConcretizationError::ParseError(err)),
             Ok(parsed_expr) => concretize_expression(parsed_expr, headers, Some(globals))?,
         };
 
         Ok(Self {
             expr,
-            context: EvaluationContext::new(headers),
+            headers_index: HeadersIndex::from_headers(headers),
         })
     }
 
@@ -634,7 +746,22 @@ impl Program {
         index: usize,
         record: &ByteRecord,
     ) -> Result<DynamicValue, SpecifiedEvaluationError> {
-        eval_expression(&self.expr, Some(index), record, &self.context)
+        eval_expression(&self.expr, Some(index), record, &self.headers_index)
+    }
+
+    pub fn run_with_record_and_last_value(
+        &self,
+        index: usize,
+        record: &ByteRecord,
+        last_value: DynamicValue,
+    ) -> Result<DynamicValue, SpecifiedEvaluationError> {
+        eval_expression_with_optional_last_value(
+            &self.expr,
+            Some(index),
+            record,
+            &self.headers_index,
+            Some(last_value),
+        )
     }
 
     pub fn run_with_record_and_globals(
@@ -643,7 +770,13 @@ impl Program {
         record: &ByteRecord,
         globals: &GlobalVariables,
     ) -> Result<DynamicValue, SpecifiedEvaluationError> {
-        eval_expression_with_globals(&self.expr, Some(index), record, &self.context, globals)
+        eval_expression_with_globals(
+            &self.expr,
+            Some(index),
+            record,
+            &self.headers_index,
+            globals,
+        )
     }
 
     pub fn generate_key(
@@ -821,14 +954,6 @@ mod tests {
         );
     }
 
-    #[test]
-    fn test_coalesce() {
-        assert_eq!(
-            eval_code("coalesce(null, false, 'test')"),
-            Ok(DynamicValue::from("test"))
-        );
-    }
-
     #[test]
     fn test_bool() {
         assert_eq!(eval_code("not(true)"), Ok(DynamicValue::from(false)));
@@ -848,7 +973,11 @@ mod tests {
     fn test_pathjoin() {
         assert_eq!(
             eval_code("pathjoin('one', 'two', 'three')"),
-            Ok(DynamicValue::from("one/two/three"))
+            Ok(DynamicValue::from(if cfg!(windows) {
+                "one\\two\\three"
+            } else {
+                "one/two/three"
+            }))
         );
     }
 
@@ -1043,7 +1172,7 @@ mod tests {
         assert_eq!(
             eval_code("col('surname', 1)"),
             Err(RunError::Prepare(ConcretizationError::ColumnNotFound(
-                ColumIndexationBy::NameAndNth(("surname".to_string(), 1))
+                ColumIndexationBy::NameAndNth("surname".to_string(), 1)
             )))
         );
         assert_eq!(eval_code("col(concat('sur', 'name'))"), Ok(b("SMITH")));
