@@ -1,5 +1,7 @@
 // En tant que chef, je m'engage à ce que nous ne nous
 // fassions pas *tous* tuer.
+use std::fmt;
+
 use lazy_static::lazy_static;
 use pest::{
     iterators::{Pair, Pairs},
@@ -39,9 +41,8 @@ impl Rule {
             Self::pow => "pow",
             Self::rem => "mod",
             Self::concat => "concat",
-            // NOTE: and & or operators need to be resolved using if statements
-            // Self::and => "and",
-            // Self::or => "or",
+            Self::and => "and",
+            Self::or => "or",
             Self::not => "not",
             Self::neg => "neg",
 
@@ -105,7 +106,8 @@ fn build_string(pair: Pair<Rule>) -> String {
         match inner.as_rule() {
             Rule::raw_double_quoted_string
             | Rule::raw_single_quoted_string
-            | Rule::raw_regex_string => {
+            | Rule::raw_regex_string
+            | Rule::raw_backtick_quoted_string => {
                 string.push_str(inner.as_str());
             }
             Rule::regex_flag => break,
@@ -121,9 +123,28 @@ fn build_string(pair: Pair<Rule>) -> String {
                             "\\" => '\\',
                             "\"" => '"',
                             "'" => '\'',
+                            "`" => '`',
+                            "0" => '\0',
                             _ => unreachable!(),
                         });
                     }
+                    Rule::byte => {
+                        string.push(char::from(
+                            u8::from_str_radix(&inner.as_str()[1..], 16).unwrap(),
+                        ));
+                    }
+                    Rule::unicode => {
+                        string.push(
+                            char::from_u32(
+                                u32::from_str_radix(
+                                    &inner.as_str()[2..inner.as_str().len() - 1],
+                                    16,
+                                )
+                                .unwrap(),
+                            )
+                            .unwrap(),
+                        );
+                    }
                     _ => unreachable!(),
                 }
             }
@@ -201,22 +222,22 @@ fn pratt_parse(pairs: Pairs<Rule>) -> Result<Expr, String> {
                 }
                 Rule::full_slice => {
                     let mut pairs = primary.into_inner();
-                    let start = parse_int(pairs.next().unwrap())?;
-                    let end = parse_int(pairs.next().unwrap())?;
+                    let start = pratt_parse(pairs.next().unwrap().into_inner())?;
+                    let end = pratt_parse(pairs.next().unwrap().into_inner())?;
 
-                    Expr::Slice(Slice::Closed(start, end))
+                    Expr::Slice(Slice::Closed(Box::new(start), Box::new(end)))
                 }
                 Rule::start_slice => {
                     let mut pairs = primary.into_inner();
-                    let start = parse_int(pairs.next().unwrap())?;
+                    let start = pratt_parse(pairs.next().unwrap().into_inner())?;
 
-                    Expr::Slice(Slice::From(start))
+                    Expr::Slice(Slice::From(Box::new(start)))
                 }
                 Rule::end_slice => {
                     let mut pairs = primary.into_inner();
-                    let end = parse_int(pairs.next().unwrap())?;
+                    let end = pratt_parse(pairs.next().unwrap().into_inner())?;
 
-                    Expr::Slice(Slice::To(end))
+                    Expr::Slice(Slice::To(Box::new(end)))
                 }
                 Rule::string => Expr::Str(build_string(primary)),
                 Rule::binary_string => Expr::BStr(build_string(primary).into_bytes()),
@@ -293,10 +314,10 @@ fn pratt_parse(pairs: Pairs<Rule>) -> Result<Expr, String> {
                         "slice",
                         match slice {
                             Slice::Closed(start, end) => {
-                                vec![lhs?, Expr::Int(start), Expr::Int(end)]
+                                vec![lhs?, *start, *end]
                             }
-                            Slice::From(start) => vec![lhs?, Expr::Int(start)],
-                            Slice::To(end) => vec![lhs?, Expr::Int(0), Expr::Int(end)],
+                            Slice::From(start) => vec![lhs?, *start],
+                            Slice::To(end) => vec![lhs?, Expr::Int(0), *end],
                             Slice::Full => unreachable!(),
                         },
                     )),
@@ -310,33 +331,20 @@ fn pratt_parse(pairs: Pairs<Rule>) -> Result<Expr, String> {
                         Some(_) => Expr::Func(FunctionCall::new(&name, vec![lhs?])),
                     },
                     mut rest => {
-                        rest.fill_underscore(&lhs?);
-                        rest
+                        let mut counter: usize = 0;
+                        rest.count_underscores(&mut counter);
+
+                        if counter == 0 {
+                            rest
+                        } else if counter == 1 {
+                            rest.fill_underscore(&lhs?);
+                            rest
+                        } else {
+                            Expr::Pipeline(vec![lhs?, rest])
+                        }
                     }
                 },
 
-                // Short-circuiting and
-                Rule::and => {
-                    let lhs_res = lhs?;
-
-                    // a && b => if(a, b, a)
-                    Expr::Func(FunctionCall::new(
-                        "if",
-                        vec![lhs_res.clone(), rhs?, lhs_res],
-                    ))
-                }
-
-                // Short-circuiting or
-                Rule::or => {
-                    let lhs_res = lhs?;
-
-                    // a || b => if(a, a, b)
-                    Expr::Func(FunctionCall::new(
-                        "if",
-                        vec![lhs_res.clone(), lhs_res, rhs?],
-                    ))
-                }
-
                 // Access & call
                 Rule::point => match rhs? {
                     Expr::Identifier(identifier, _) => {
@@ -410,6 +418,7 @@ pub enum Slice<T> {
 #[derive(Debug, PartialEq, Clone)]
 pub enum Expr {
     Func(FunctionCall),
+    Pipeline(Vec<Expr>),
     Lambda(Vec<String>, Box<Expr>),
     LambdaBinding(String),
     Int(i64),
@@ -420,7 +429,7 @@ pub enum Expr {
     List(Vec<Expr>),
     Map(Vec<(String, Expr)>),
     Regex(String, bool),
-    Slice(Slice<i64>),
+    Slice(Slice<Box<Expr>>),
     StarSlice(Slice<DynamicValue>),
     Bool(bool),
     Underscore,
@@ -483,6 +492,30 @@ impl Expr {
         };
     }
 
+    fn count_underscores(&self, counter: &mut usize) {
+        match self {
+            Expr::Func(call) => {
+                for (_, arg) in call.args.iter() {
+                    arg.count_underscores(counter);
+                }
+            }
+            Expr::List(list) => {
+                for item in list.iter() {
+                    item.count_underscores(counter);
+                }
+            }
+            Expr::Map(map) => {
+                for (_, value) in map.iter() {
+                    value.count_underscores(counter);
+                }
+            }
+            Expr::Underscore => {
+                *counter += 1;
+            }
+            _ => (),
+        }
+    }
+
     fn fill_underscore(&mut self, with: &Expr) {
         match self {
             Expr::Func(call) => {
@@ -552,6 +585,15 @@ pub enum ParseError {
     Custom(String),
 }
 
+impl fmt::Display for ParseError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::Pest(err) => write!(f, "{}", err),
+            Self::Custom(msg) => write!(f, "{}", msg),
+        }
+    }
+}
+
 impl From<pest::error::Error<Rule>> for ParseError {
     fn from(value: pest::error::Error<Rule>) -> Self {
         Self::Pest(Box::new(value))
@@ -1127,7 +1169,10 @@ mod tests {
         // Double underscore
         assert_eq!(
             parse_expression("count | add(_, _)"),
-            Ok(func("add", vec![id("count"), id("count")]))
+            Ok(Expr::Pipeline(vec![
+                id("count"),
+                func("add", vec![Expr::Underscore, Expr::Underscore])
+            ]))
         );
 
         // Nested underscores
