@@ -1,11 +1,12 @@
 use std::cmp::Ordering;
-use std::collections::HashSet;
 use std::convert::TryFrom;
 use std::fmt;
 use std::iter::repeat;
 use std::ops;
 use std::str::FromStr;
 
+use crate::collections::HashSet;
+
 #[derive(Clone, Deserialize)]
 #[serde(try_from = "String")]
 pub struct SelectColumns {
@@ -135,6 +136,12 @@ impl TryFrom<String> for SelectColumns {
     }
 }
 
+impl Default for SelectColumns {
+    fn default() -> Self {
+        Self::parse("").unwrap()
+    }
+}
+
 struct SelectorParser {
     chars: Vec<char>,
     pos: usize,
@@ -568,25 +575,35 @@ impl Selection {
         }
     }
 
-    pub fn select<'a, 'b>(&'a self, row: &'b csv::ByteRecord) -> impl Iterator<Item = &'b [u8]>
-    where
-        'a: 'b,
-    {
-        self.iter().map(move |i| &row[*i])
+    pub fn insert(&mut self, index: usize, element: usize) {
+        self.0.insert(index, element);
     }
 
-    pub fn select_string_record<'a, 'b>(
+    pub fn select<'a, 'b, T: 'b + ?Sized>(
         &'a self,
-        row: &'b csv::StringRecord,
-    ) -> impl Iterator<Item = &'b str>
+        row: &'b impl ops::Index<usize, Output = T>,
+    ) -> impl Iterator<Item = &'b T>
     where
         'a: 'b,
     {
-        self.iter().map(move |i| &row[*i])
+        self.iter().map(|i| &row[*i])
     }
 
     pub fn collect(&self, row: &csv::ByteRecord) -> Vec<Vec<u8>> {
-        self.select(row).map(|f| f.to_vec()).collect()
+        self.select(row).map(|f| f.into()).collect()
+    }
+
+    pub fn dedup(&mut self) {
+        let mut new = Vec::new();
+        let mut seen = HashSet::new();
+
+        for i in self.0.iter().copied() {
+            if seen.insert(i) {
+                new.push(i);
+            }
+        }
+
+        self.0 = new;
     }
 
     pub fn sort_and_dedup(&mut self) {
@@ -613,6 +630,13 @@ impl Selection {
         m
     }
 
+    pub fn mask(&self, alignment: usize) -> Vec<bool> {
+        self.indexed_mask(alignment)
+            .into_iter()
+            .map(|o| o.is_some())
+            .collect()
+    }
+
     pub fn len(&self) -> usize {
         self.0.len()
     }
@@ -628,6 +652,20 @@ impl Selection {
     pub fn subtract(&mut self, other: &Self) {
         self.0.retain(|i| !other.contains(*i));
     }
+
+    pub fn inverse(&self, alignment: usize) -> Self {
+        let mask = self.mask(alignment);
+
+        let mut indices = Vec::new();
+
+        for (i, is_selected) in mask.into_iter().enumerate() {
+            if !is_selected {
+                indices.push(i);
+            }
+        }
+
+        Self(indices)
+    }
 }
 
 impl ops::Deref for Selection {
