@@ -5,11 +5,9 @@ use std::fs;
 use std::io;
 use std::num::NonZeroUsize;
 use std::ops::Deref;
-use std::path::{Path, PathBuf};
+use std::path::Path;
 use std::process::Command;
 use std::str;
-use std::thread;
-use std::time;
 
 use colored::{Color, ColoredString, Colorize, Styles};
 use deepsize::DeepSizeOf;
@@ -50,11 +48,12 @@ pub fn version() -> String {
 }
 
 lazy_static! {
-    static ref FLAG_REGEX: Regex = Regex::new(r"([\s,/\(])(--?[A-Za-z][\w\-]*)").unwrap();
+    static ref FLAG_REGEX: Regex = Regex::new(r"([\s,/\(])(--?[A-Za-z§][\w\-]*)").unwrap();
     static ref SECTION_REGEX: Regex = Regex::new("(?im)^.*(?:usage|options?):|---+").unwrap();
-    static ref DIMMED_REGEX: Regex =
-        Regex::new(r"\[--\]|\[?<[\w|\-]+>(?:\.{3})?\]?|\[[\w\s:§|]+\]|\s+[\$>][^\n]+|\*[^*\n]+\*")
-            .unwrap();
+    static ref DIMMED_REGEX: Regex = Regex::new(
+        r"\[--\]|\[?<[\w|\-]+>(?:\.{3})?\]?|\[[\w\s:§|\-.]+\]|\s+[\$>][^\n]+|\*[^*\n]+\*"
+    )
+    .unwrap();
     static ref QUOTE_REGEX: Regex = Regex::new(r#"(?m)"[^"\n]+"|'[^'\n]+'|`[^`\n]+`"#).unwrap();
     static ref MAIN_SECTION_REGEX: Regex = Regex::new("(?m)^#+.+").unwrap();
     static ref MAIN_COMMAND_REGEX: Regex = Regex::new(r"(?m)^\s{4}[\w\-]+").unwrap();
@@ -142,32 +141,6 @@ pub fn errif_greater_one_stdin(inps: &[Config]) -> Result<(), String> {
     Ok(())
 }
 
-pub fn num_of_chunks(nitems: usize, chunk_size: usize) -> usize {
-    if chunk_size == 0 {
-        return nitems;
-    }
-    let mut n = nitems / chunk_size;
-    if nitems % chunk_size != 0 {
-        n += 1;
-    }
-    n
-}
-
-pub fn last_modified(md: &fs::Metadata) -> u64 {
-    use filetime::FileTime;
-    FileTime::from_last_modification_time(md).seconds() as u64
-}
-
-pub fn idx_path(csv_path: &Path) -> PathBuf {
-    let mut p = csv_path
-        .to_path_buf()
-        .into_os_string()
-        .into_string()
-        .unwrap();
-    p.push_str(".idx");
-    PathBuf::from(&p)
-}
-
 pub type Idx = Option<usize>;
 
 pub fn range(start: Idx, end: Idx, len: Idx, index: Idx) -> Result<(usize, usize), String> {
@@ -197,27 +170,6 @@ pub fn range(start: Idx, end: Idx, len: Idx, index: Idx) -> Result<(usize, usize
     }
 }
 
-/// Create a directory recursively, avoiding the race conditons fixed by
-/// https://github.com/rust-lang/rust/pull/39799.
-fn create_dir_all_threadsafe(path: &Path) -> io::Result<()> {
-    // Try 20 times. This shouldn't theoretically need to be any larger
-    // than the number of nested directories we need to create.
-    for _ in 0..20 {
-        match fs::create_dir_all(path) {
-            // This happens if a directory in `path` doesn't exist when we
-            // test for it, and another thread creates it before we can.
-            Err(ref err) if err.kind() == io::ErrorKind::AlreadyExists => {}
-            other => return other,
-        }
-        // We probably don't need to sleep at all, because the intermediate
-        // directory is already created.  But let's attempt to back off a
-        // bit and let the other thread finish.
-        thread::sleep(time::Duration::from_millis(25));
-    }
-    // Try one last time, returning whatever happens.
-    fs::create_dir_all(path)
-}
-
 /// Represents a filename template of the form `"{}.csv"`, where `"{}"` is
 /// the splace to insert the part of the filename generated by `xan`.
 #[derive(Clone, Debug, Deserialize)]
@@ -252,7 +204,7 @@ impl FilenameTemplate {
             // We may be called concurrently, especially by parallel `xan
             // split`, so be careful to avoid the `create_dir_all` race
             // condition.
-            create_dir_all_threadsafe(parent)?;
+            fs::create_dir_all(parent)?;
         }
         let spath = Some(full_path.display().to_string());
         Config::new(&spath).writer_with_options(
@@ -415,13 +367,19 @@ pub enum ColorOrStyles {
     Styles(Styles),
 }
 
+impl ColorOrStyles {
+    pub fn is_green(&self) -> bool {
+        matches!(self, Self::Color(color) if color == &Color::Green)
+    }
+}
+
 pub fn colorizer_by_type(string: &str) -> ColorOrStyles {
     match string {
         "true" | "TRUE" | "True" | "false" | "FALSE" | "False" | "yes" | "no" => {
             return ColorOrStyles::Color(Color::Cyan)
         }
-        "null" | "na" | "NA" | "None" | "n/a" | "N/A" | "nan" | "NaN" | "<empty>" | "<null>"
-        | "<rest>" | "." | "-" => return ColorOrStyles::Styles(Styles::Dimmed),
+        "NULL" | "null" | "na" | "NA" | "None" | "n/a" | "N/A" | "nan" | "NaN" | "<empty>"
+        | "<null>" | "<rest>" | "." | "-" => return ColorOrStyles::Styles(Styles::Dimmed),
         _ => (),
     };
 
@@ -552,15 +510,20 @@ pub fn unicode_aware_pad<'a>(
         return Cow::Borrowed(string);
     }
 
-    let mut padded = String::new();
-    let padding = padding.repeat(width - string_width);
+    let mut padded = String::with_capacity(width);
 
     if left {
-        padded.push_str(&padding);
+        for _ in 0..(width - string_width) {
+            padded.push_str(padding);
+        }
+
         padded.push_str(string);
     } else {
         padded.push_str(string);
-        padded.push_str(&padding);
+
+        for _ in 0..(width - string_width) {
+            padded.push_str(padding);
+        }
     }
 
     Cow::Owned(padded)
@@ -854,12 +817,6 @@ impl<T: Deref<Target = str>, I: Iterator<Item = T>> JoinIteratorExt for I {
     }
 }
 
-pub fn bytes_cursor_from_read<R: io::Read>(source: &mut R) -> io::Result<io::Cursor<Vec<u8>>> {
-    let mut bytes = Vec::<u8>::new();
-    source.read_to_end(&mut bytes)?;
-    Ok(io::Cursor::new(bytes))
-}
-
 // A custom implementation de/serializing ext-sort chunks as CSV
 pub struct DeepSizedByteRecord(pub csv::ByteRecord);
 
@@ -875,8 +832,7 @@ impl DeepSizedByteRecord {
 
 impl DeepSizeOf for DeepSizedByteRecord {
     fn deep_size_of(&self) -> usize {
-        // Good enough approximation...
-        self.0.as_slice().len() + (self.0.len().saturating_sub(1))
+        std::mem::size_of::<csv::ByteRecord>() + self.0.as_slice().len()
     }
 
     fn deep_size_of_children(&self, _context: &mut deepsize::Context) -> usize {
@@ -894,7 +850,9 @@ impl ExternalChunk<DeepSizedByteRecord> for CsvExternalChunk {
 
     fn new(reader: io::Take<io::BufReader<fs::File>>) -> Self {
         CsvExternalChunk {
-            reader: csv::Reader::from_reader(reader),
+            reader: csv::ReaderBuilder::new()
+                .has_headers(false)
+                .from_reader(reader),
         }
     }
 
