@@ -531,8 +531,8 @@ fn agg_correlation() {
 }
 
 #[test]
-fn agg_cols() {
-    let wrk = Workdir::new("agg_cols");
+fn agg_along_rows() {
+    let wrk = Workdir::new("agg_along_rows");
     wrk.create(
         "data.csv",
         vec![
@@ -543,8 +543,8 @@ fn agg_cols() {
     );
 
     let mut cmd = wrk.command("agg");
-    cmd.arg("mean(cell) as mean")
-        .args(["--cols", "a,b"])
+    cmd.arg("mean(_) as mean")
+        .args(["--along-rows", "a,b"])
         .arg("data.csv");
 
     let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
@@ -555,3 +555,47 @@ fn agg_cols() {
     ];
     assert_eq!(got, expected);
 }
+
+#[test]
+fn agg_along_cols() {
+    let wrk = Workdir::new("agg_along_cols");
+    wrk.create(
+        "data.csv",
+        vec![
+            svec!["name", "a", "b"],
+            svec!["john", "3", "5"],
+            svec!["lucy", "6", "1"],
+        ],
+    );
+
+    let mut cmd = wrk.command("agg");
+    cmd.arg("sum(_)")
+        .args(["--along-cols", "a,b"])
+        .arg("data.csv");
+
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![svec!["a", "b"], svec!["9", "6"]];
+    assert_eq!(got, expected);
+}
+
+#[test]
+fn agg_along_matrix() {
+    let wrk = Workdir::new("agg_along_matrix");
+    wrk.create(
+        "data.csv",
+        vec![
+            svec!["name", "a", "b"],
+            svec!["john", "3", "5"],
+            svec!["lucy", "6", "1"],
+        ],
+    );
+
+    let mut cmd = wrk.command("agg");
+    cmd.arg("sum(_) as total")
+        .args(["--along-matrix", "a,b"])
+        .arg("data.csv");
+
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![svec!["total"], svec!["15"]];
+    assert_eq!(got, expected);
+}
