@@ -141,12 +141,27 @@ fn search_invert_match_no_headers() {
 fn search_count() {
     let wrk = Workdir::new("search_count");
     wrk.create("data.csv", data(false));
+
     let mut cmd = wrk.command("search");
     cmd.arg("-r")
         .arg("foo")
         .arg("data.csv")
         .args(["--count", "matches"]);
 
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![
+        svec!["foobar", "barfoo", "matches"],
+        svec!["barfoo", "foobar", "2"],
+    ];
+    assert_eq!(got, expected);
+
+    let mut cmd = wrk.command("search");
+    cmd.arg("-r")
+        .arg("foo")
+        .arg("--left")
+        .arg("data.csv")
+        .args(["--count", "matches"]);
+
     let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
     let expected = vec![
         svec!["foobar", "barfoo", "matches"],
@@ -499,6 +514,24 @@ fn search_count_patterns_regex() {
         .args(["--count", "matches"])
         .arg("data.csv");
 
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![
+        svec!["text", "matches"],
+        svec!["Lucy went to school with John.", "2"],
+        svec!["john is dead. poor john", "2"],
+        svec!["Lucy in the sky with diamonds", "1"],
+    ];
+    assert_eq!(got, expected);
+
+    let mut cmd = wrk.command("search");
+    cmd.arg("-r")
+        .arg("-i")
+        .arg("--left")
+        .args(["--patterns", "patterns.csv"])
+        .args(["--pattern-column", "pattern"])
+        .args(["--count", "matches"])
+        .arg("data.csv");
+
     let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
     let expected = vec![
         svec!["text", "matches"],
@@ -644,16 +677,41 @@ fn search_replace() {
     let wrk = Workdir::new("search_replace");
 
     wrk.create(
-        "data.csv",
+        "data1.csv",
         vec![svec!["number"], svec!["3,4"], svec!["2"], svec!["10,7"]],
     );
 
+    wrk.create(
+        "data2.csv",
+        vec![
+            svec!["id", "number"],
+            svec!["1", "3,4"],
+            svec!["2", "2"],
+            svec!["3", "10,7"],
+        ],
+    );
+
     let mut cmd = wrk.command("search");
-    cmd.arg(",").args(["--replace", "."]).arg("data.csv");
+    cmd.arg(",").args(["--replace", "."]).arg("data1.csv");
 
     let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
     let expected = vec![svec!["number"], svec!["3.4"], svec!["2"], svec!["10.7"]];
     assert_eq!(got, expected);
+
+    let mut cmd = wrk.command("search");
+    cmd.arg(",")
+        .args(["--replace", "."])
+        .args(["-s", "number"])
+        .arg("data2.csv");
+
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![
+        svec!["id", "number"],
+        svec!["1", "3.4"],
+        svec!["2", "2"],
+        svec!["3", "10.7"],
+    ];
+    assert_eq!(got, expected);
 }
 
 #[test]
@@ -906,3 +964,160 @@ fn search_patterns_url_replace() {
     ];
     assert_eq!(got, expected);
 }
+
+#[test]
+fn search_patterns_breakdown() {
+    let wrk = Workdir::new("search_patterns_breakdown");
+
+    wrk.create(
+        "patterns.csv",
+        vec![
+            svec!["article", "name"],
+            svec!["le", "LE"],
+            svec!["la", "LA"],
+        ],
+    );
+
+    wrk.create(
+        "data.csv",
+        vec![
+            svec!["text"],
+            svec!["le chien mange le fromage"],
+            svec!["no match"],
+            svec!["le chien mange la souris"],
+        ],
+    );
+
+    let mut cmd = wrk.command("search");
+    cmd.args(["--patterns", "patterns.csv"])
+        .args(["--pattern-column", "article"])
+        .arg("-B")
+        .arg("data.csv");
+
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![
+        svec!["text", "le", "la"],
+        svec!["le chien mange le fromage", "2", "0"],
+        svec!["le chien mange la souris", "1", "1"],
+    ];
+    assert_eq!(got, expected);
+
+    let mut cmd = wrk.command("search");
+    cmd.args(["--patterns", "patterns.csv"])
+        .args(["--pattern-column", "article"])
+        .arg("-B")
+        .args(["--name-column", "name"])
+        .arg("data.csv");
+
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![
+        svec!["text", "LE", "LA"],
+        svec!["le chien mange le fromage", "2", "0"],
+        svec!["le chien mange la souris", "1", "1"],
+    ];
+    assert_eq!(got, expected);
+
+    let mut cmd = wrk.command("search");
+    cmd.args(["--patterns", "patterns.csv"])
+        .args(["--pattern-column", "article"])
+        .arg("-B")
+        .arg("--left")
+        .args(["--name-column", "name"])
+        .arg("data.csv");
+
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![
+        svec!["text", "LE", "LA"],
+        svec!["le chien mange le fromage", "2", "0"],
+        svec!["no match", "0", "0"],
+        svec!["le chien mange la souris", "1", "1"],
+    ];
+    assert_eq!(got, expected);
+}
+
+#[test]
+fn search_patterns_unique_matches() {
+    let wrk = Workdir::new("search_patterns_unique_matches");
+
+    wrk.create(
+        "patterns.csv",
+        vec![
+            svec!["article", "name"],
+            svec!["la", "LA"],
+            svec!["le", "LE"],
+        ],
+    );
+
+    wrk.create(
+        "data.csv",
+        vec![
+            svec!["text"],
+            svec!["le chien mange le fromage"],
+            svec!["le chien mange la souris"],
+            svec!["coucou"],
+        ],
+    );
+
+    let mut cmd = wrk.command("search");
+    cmd.args(["--patterns", "patterns.csv"])
+        .args(["--pattern-column", "article"])
+        .args(["-U", "matches"])
+        .arg("data.csv");
+
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![
+        svec!["text", "matches"],
+        svec!["le chien mange le fromage", "le"],
+        svec!["le chien mange la souris", "la|le"],
+    ];
+    assert_eq!(got, expected);
+
+    // --name-column
+    let mut cmd = wrk.command("search");
+    cmd.args(["--patterns", "patterns.csv"])
+        .args(["--pattern-column", "article"])
+        .args(["--name-column", "name"])
+        .args(["-U", "matches"])
+        .arg("data.csv");
+
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![
+        svec!["text", "matches"],
+        svec!["le chien mange le fromage", "LE"],
+        svec!["le chien mange la souris", "LA|LE"],
+    ];
+    assert_eq!(got, expected);
+
+    // --sep
+    let mut cmd = wrk.command("search");
+    cmd.args(["--patterns", "patterns.csv"])
+        .args(["--pattern-column", "article"])
+        .args(["-U", "matches"])
+        .args(["--sep", "§"])
+        .arg("data.csv");
+
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![
+        svec!["text", "matches"],
+        svec!["le chien mange le fromage", "le"],
+        svec!["le chien mange la souris", "la§le"],
+    ];
+    assert_eq!(got, expected);
+
+    // --left
+    let mut cmd = wrk.command("search");
+    cmd.args(["--patterns", "patterns.csv"])
+        .args(["--pattern-column", "article"])
+        .args(["-U", "matches"])
+        .arg("--left")
+        .arg("data.csv");
+
+    let got: Vec<Vec<String>> = wrk.read_stdout(&mut cmd);
+    let expected = vec![
+        svec!["text", "matches"],
+        svec!["le chien mange le fromage", "le"],
+        svec!["le chien mange la souris", "la|le"],
+        svec!["coucou", ""],
+    ];
+    assert_eq!(got, expected);
+}
